# flake8: noqa
"""
Absfuyu: Lunar calendar
-----------------------
Convert to lunar calendar

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)

Source:
-------
Astronomical algorithms from the book "Astronomical Algorithms" by Jean Meeus, 1998
https://www.informatik.uni-leipzig.de/~duc/amlich/AL.py
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["LunarCalendar"]


# Library
# ---------------------------------------------------------------------------
import math
from datetime import date, datetime


# Class
# ---------------------------------------------------------------------------
class LunarCalendar:
    """Lunar Calendar"""

    def __init__(
        self,
        year: int,
        month: int,
        day: int,
        *,
        time_zone: int = 7,
        lunar_leap: bool = False,
    ) -> None:
        """
        :param time_zone: Time zone
        :param lunar_leap: Do not use this
        """
        self.date = date(year, month, day)
        self.time_zone = time_zone
        self._lunar_leap = lunar_leap

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.date}|lunar_leap={self._lunar_leap})"

    def __repr__(self) -> str:
        return self.__str__()

    def _julian_day_from_date(
        self,
        *,
        overwrite_year: int | None = None,
        overwrite_month: int | None = None,
        overwrite_day: int | None = None,
    ) -> int:
        """
        Compute the (integral) Julian day number of `self.date`

        i.e., the number of days between 1/1/4713 BC (Julian calendar) and `self.date`.
        """
        day = self.date.day if overwrite_day is None else overwrite_day
        month = self.date.month if overwrite_month is None else overwrite_month
        year = self.date.year if overwrite_year is None else overwrite_year
        a = int((14 - month) / 12.0)
        y = year + 4800 - a
        m = month + 12 * a - 3
        jd = (
            day
            + int((153 * m + 2) / 5.0)
            + 365 * y
            + int(y / 4.0)
            - int(y / 100.0)
            + int(y / 400.0)
            - 32045
        )
        if jd < 2299161:
            jd = day + int((153 * m + 2) / 5.0) + 365 * y + int(y / 4.0) - 32083
        return jd

    @staticmethod
    def _julian_day_to_date(julian_day: int) -> date:
        """
        Convert a Julian day number to `datetime.date`

        julian_day: Julian day
        """
        if julian_day > 2299160:
            ## After 5/10/1582, Gregorian calendar
            a = julian_day + 32044
            b = int((4 * a + 3) / 146097.0)
            c = a - int((b * 146097) / 4.0)
        else:
            b = 0
            c = julian_day + 32082

        d = int((4 * c + 3) / 1461.0)
        e = c - int((1461 * d) / 4.0)
        m = int((5 * e + 2) / 153.0)
        day = e - int((153 * m + 2) / 5.0) + 1
        month = m + 3 - 12 * int(m / 10.0)
        year = b * 100 + d - 4800 + int(m / 10.0)
        return date(year, month, day)

    @staticmethod
    def _new_moon(k: int) -> float:
        """
        Compute the time of the k-th new moon after the new moon of `1/1/1900 13:52 UCT`
        measured as the number of days since `1/1/4713 BC noon UCT`,
        e.g., `2451545.125 is 1/1/2000 15:00 UTC`.

        Example:
        `2415079.9758617813` for `k=2` or `2414961.935157746` for `k=-2`
        """
        ## Time in Julian centuries from 1900 January 0.5
        T = k / 1236.85
        T2 = math.pow(T, 2)
        T3 = math.pow(T, 3)
        dr = math.pi / 180.0
        Jd1 = 2415020.75933 + 29.53058868 * k + 0.0001178 * T2 - 0.000000155 * T3
        Jd1 = Jd1 + 0.00033 * math.sin((166.56 + 132.87 * T - 0.009173 * T2) * dr)
        ## Mean new moon
        M = 359.2242 + 29.10535608 * k - 0.0000333 * T2 - 0.00000347 * T3
        ## Sun's mean anomaly
        Mpr = 306.0253 + 385.81691806 * k + 0.0107306 * T2 + 0.00001236 * T3
        ## Moon's mean anomaly
        F = 21.2964 + 390.67050646 * k - 0.0016528 * T2 - 0.00000239 * T3
        ## Moon's argument of latitude
        C1 = (0.1734 - 0.000393 * T) * math.sin(M * dr) + 0.0021 * math.sin(2 * dr * M)
        C1 = C1 - 0.4068 * math.sin(Mpr * dr) + 0.0161 * math.sin(dr * 2 * Mpr)
        C1 = C1 - 0.0004 * math.sin(dr * 3 * Mpr)
        C1 = C1 + 0.0104 * math.sin(dr * 2 * F) - 0.0051 * math.sin(dr * (M + Mpr))
        C1 = (
            C1 - 0.0074 * math.sin(dr * (M - Mpr)) + 0.0004 * math.sin(dr * (2 * F + M))
        )
        C1 = (
            C1
            - 0.0004 * math.sin(dr * (2 * F - M))
            - 0.0006 * math.sin(dr * (2 * F + Mpr))
        )
        C1 = (
            C1
            + 0.0010 * math.sin(dr * (2 * F - Mpr))
            + 0.0005 * math.sin(dr * (2 * Mpr + M))
        )
        if T < -11:
            deltat = (
                0.001
                + 0.000839 * T
                + 0.0002261 * T2
                - 0.00000845 * T3
                - 0.000000081 * T * T3
            )
        else:
            deltat = -0.000278 + 0.000265 * T + 0.000262 * T2
        JdNew = Jd1 + C1 - deltat
        return JdNew

    @staticmethod
    def _sun_longitude(julian_day_noon: float) -> float:
        """
        Compute the longitude of the sun at any time.

        julian_day_noon: The number of days since `1/1/4713 BC noon`.
        """
        T = (julian_day_noon - 2451545.0) / 36525.0
        # Time in Julian centuries
        # from 2000-01-01 12:00:00 GMT
        T2 = math.pow(T, 2)
        T3 = math.pow(T, 3)
        dr = math.pi / 180.0  # degree to radian
        M = 357.52910 + 35999.05030 * T - 0.0001559 * T2 - 0.00000048 * T3
        # mean anomaly, degree
        L0 = 280.46645 + 36000.76983 * T + 0.0003032 * T2
        # mean longitude, degree
        DL = (1.914600 - 0.004817 * T - 0.000014 * T2) * math.sin(dr * M)
        DL += (0.019993 - 0.000101 * T) * math.sin(dr * 2 * M) + 0.000290 * math.sin(
            dr * 3 * M
        )
        L = L0 + DL  # true longitude, degree
        L = L * dr
        L = L - math.pi * 2 * (int(L / (math.pi * 2)))
        # Normalize to (0, 2*math.pi)
        return L

    def _get_sun_longitude(self, day_number: int) -> int:
        """
        Compute sun position at midnight of the day with the given Julian day number.

        The time zone if the time difference between local time and UTC: 7.0 for UTC+7:00.

        The function returns a number between `0` and `11`.

        From the day after March equinox and the 1st major term after March equinox, 0 is returned. After that, return 1, 2, 3 ...
        """
        return int(
            self._sun_longitude(day_number - 0.5 - self.time_zone / 24.0) / math.pi * 6
        )

    def _get_new_moon_day(self, k: int) -> int:
        """
        Compute the day of the k-th new moon in the given time zone.

        The time zone if the time difference between local time and UTC: 7.0 for UTC+7:00.
        """
        return int(self._new_moon(k) + 0.5 + self.time_zone / 24.0)

    def _get_lunar_month_11(self, *, overwrite_year: int | None = None) -> int:
        """
        Find the day that starts the luner month 11
        of the given year for the given time zone.
        """
        # off = self._julian_day_from_date(overwrite_month=12, overwrite_day=31) - 2415021.076998695
        year = self.date.year if overwrite_year is None else overwrite_year
        off = (
            self._julian_day_from_date(
                overwrite_month=12, overwrite_day=31, overwrite_year=year
            )
            - 2415021.0
        )
        k = int(off / 29.530588853)
        nm = self._get_new_moon_day(k)
        sunLong = self._get_sun_longitude(nm)
        #### sun longitude at local midnight
        if sunLong >= 9:
            nm = self._get_new_moon_day(k - 1)
        return nm

    def _get_leap_month_offset(self, a11: int) -> int:
        """
        Find the index of the leap month after the month starting on the day a11.
        """
        k = int((a11 - 2415021.076998695) / 29.530588853 + 0.5)
        last = 0
        i = 1  # start with month following lunar month 11
        arc = self._get_sun_longitude(self._get_new_moon_day(k + i))
        while True:
            last = arc
            i += 1
            arc = self._get_sun_longitude(self._get_new_moon_day(k + i))
            if not (arc != last and i < 14):
                break

        # for i in range(1, 14):
        #     arc = self._get_sun_longitude(self._get_new_moon_day(k + i))
        #     if arc != last:
        #         last = arc
        #     else:
        #         break
        return i - 1

    def to_lunar(self):
        """
        Convert solar date to the corresponding lunar date.

        Returns
        -------
        LunarCalendar
            LunarCalendar instance
        """
        yy = self.date.year
        day_number = self._julian_day_from_date()
        k = int((day_number - 2415021.076998695) / 29.530588853)
        month_start = self._get_new_moon_day(k + 1)
        if month_start > day_number:
            month_start = self._get_new_moon_day(k)
        # alert(day_number + " -> " + month_start)
        a11 = self._get_lunar_month_11()
        b11 = a11
        if a11 >= month_start:
            lunar_year = yy
            a11 = self._get_lunar_month_11(overwrite_year=yy - 1)
        else:
            lunar_year = yy + 1
            b11 = self._get_lunar_month_11(overwrite_year=yy + 1)
        lunar_day = day_number - month_start + 1
        diff = int((month_start - a11) / 29.0)
        lunar_month = diff + 11
        if b11 - a11 > 365:
            leap_month_diff = self._get_leap_month_offset(a11)
            if diff >= leap_month_diff:
                lunar_month = diff + 10
                if diff == leap_month_diff:
                    self._lunar_leap = True
        if lunar_month > 12:
            lunar_month = lunar_month - 12
        if lunar_month >= 11 and diff < 4:
            lunar_year -= 1
        return self.__class__(
            lunar_year, lunar_month, lunar_day, lunar_leap=self._lunar_leap
        )

    def to_solar(self):
        """
        Convert a lunar date to the corresponding solar date.

        Returns
        -------
        LunarCalendar
            LunarCalendar instance

        date
            When unable to convert
        """
        lunarD = self.date.day
        lunarM = self.date.month
        lunarY = self.date.year
        if lunarM < 11:
            a11 = self._get_lunar_month_11(overwrite_year=lunarY - 1)
            b11 = self._get_lunar_month_11(overwrite_year=lunarY)
        else:
            a11 = self._get_lunar_month_11(overwrite_year=lunarY)
            b11 = self._get_lunar_month_11(overwrite_year=lunarY + 1)
        k = int(0.5 + (a11 - 2415021.076998695) / 29.530588853)
        off = lunarM - 11
        if off < 0:
            off += 12
        if b11 - a11 > 365:
            leapOff = self._get_leap_month_offset(a11)
            leapM = leapOff - 2
            if leapM < 0:
                leapM += 12
            if self._lunar_leap is True and lunarM != leapM:
                # logger.warning("lunar_leap is True")
                # raise ValueError()
                return date(1, 1, 1)
            elif self._lunar_leap is True or off >= leapOff:
                off += 1
        monthStart = self._get_new_moon_day(k + off)
        out = self._julian_day_to_date(monthStart + lunarD - 1)
        return self.__class__(out.year, out.month, out.day, lunar_leap=self._lunar_leap)

    def convert_all(self) -> dict:
        """
        Convert to both lunar and solar

        This method has no practical use

        :rtype: dict
        """
        return {"lunar": self.to_lunar(), "original": self, "solar": self.to_solar()}

    @classmethod
    def from_solar(cls, year: int, month: int, day: int, lunar_leap: bool = False):
        """
        Convert to lunar day from solar day

        Parameters
        ----------
        year : int
            Year (in range [1, 9999])

        month : int
            Month (in range [1, 12])

        day : int
            Day (in range [1, 31])

        lunar_leap : bool
            Is leap (Default: ``False``)

        Returns
        -------
        LunarCalendar
            LunarCalendar instance
        """
        return cls(year, month, day, lunar_leap=lunar_leap).to_lunar()

    @classmethod
    def from_datetime(cls, datetime_object: date | datetime):
        """
        Convert from ``datetime.datetime`` object

        Parameters
        ----------
        datetime_object : date | datetime
            ``datetime.datetime`` object

        Returns
        -------
        LunarCalendar
            LunarCalendar instance
        """
        return cls(datetime_object.year, datetime_object.month, datetime_object.day)

    @classmethod
    def now(cls):
        """
        Get lunar calendar for ``datetime.datetime.now()``

        Returns
        -------
        LunarCalendar
            LunarCalendar instance
        """
        today = datetime.now().date()
        return cls(today.year, today.month, today.day).to_lunar()



"""
Absfuyu: Shutdownizer
---------------------
This shutdowns

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["ShutDownizer", "ShutdownEngine"]


# Library
# ---------------------------------------------------------------------------
import os
import subprocess
import sys
from abc import ABC, abstractmethod
from datetime import datetime, time, timedelta
from pathlib import Path
from typing import Annotated

try:
    from typing import override  # type: ignore
except ImportError:
    from absfuyu.core.decorator import dummy_decorator as override

from absfuyu.core import BaseClass, versionadded, versionchanged
from absfuyu.logger import logger

# TODO: Schedule shutdown, random time shutdown, test


# Class
# ---------------------------------------------------------------------------
@versionadded("4.2.0")
class ShutDownizer(BaseClass):
    """
    ShutDownizer

    Shutdown tool because why not
    """

    __slots__ = ("os", "engine")

    def __init__(self) -> None:
        self.os: str = sys.platform
        logger.debug(f"Current OS: {self.os}")

        if self.os in ["win32", "cygwin"]:  # Windows
            self.engine = ShutdownEngineWin()  # type: ignore
        elif self.os == "darwin":  # MacOS
            self.engine = ShutdownEngineMac()  # type: ignore
        elif self.os == "linux":  # Linux
            self.engine = ShutdownEngineLinux()  # type: ignore
        else:
            raise SystemError("OS not supported")

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.os})"

    def shutdown(self, *args, **kwargs) -> None:
        """Shutdown"""
        self.engine.shutdown(*args, **kwargs)

    def restart(self, *args, **kwargs) -> None:
        """Restart"""
        self.engine.restart(*args, **kwargs)

    def cancel(self) -> None:
        """Cancel"""
        self.engine.cancel()


class ShutdownEngine(ABC, BaseClass):
    """
    Abstract shutdown class for different type of OS
    """

    def __str__(self) -> str:
        return f"{self.__class__.__name__}()"

    def _execute_cmd(self, cmd: str | list) -> None:
        """Execute the cmd"""
        try:
            if isinstance(cmd, str):
                subprocess.run(cmd.split())
            elif isinstance(cmd, list):
                subprocess.run(cmd)
        except (FileNotFoundError, Exception) as e:
            logger.error(f'"{cmd}" failed to run: {e}')
            raise ValueError(f'"{cmd}" failed to run')  # noqa

    def _execute_multiple_cmds(self, cmds: list) -> None:
        if not isinstance(cmds, list):
            raise ValueError("cmds must be a <list>")
        for cmd in cmds:
            try:
                logger.debug(f"Executing: {cmd}")
                self._execute_cmd(cmd)
                break
            except Exception as e:
                logger.error(f'"{cmd}" failed to run: {e}')

    @abstractmethod
    def shutdown(self, *args, **kwargs) -> None:
        """Shutdown"""
        pass

    @abstractmethod
    def restart(self, *args, **kwargs) -> None:
        """Restart"""
        pass

    @abstractmethod
    def sleep(self, *args, **kwargs) -> None:
        """Sleep"""
        pass

    @abstractmethod
    def abort(self) -> None:
        """Abort/Cancel"""
        pass

    def cancel(self) -> None:
        """Abort/Cancel"""
        self.abort()

    def _calculate_time(
        self,
        h: Annotated[int, "positive"] = 0,
        m: Annotated[int, "positive"] = 0,
        aggregate: bool = True,
    ) -> int:
        """
        Calculate time for scheduled shutdown.

        Parameters
        ----------
        h : int, optional
            Hours to add (24h format), by default ``0``

        m : int, optional
            Minutes to add (24h format), by default ``0``

        aggregate : bool, optional
            This add hours and and minutes to `time.now()`, by default ``True``
            - ``True`` : Add hours and minutes to current time
            - ``False``: Use ``h`` and ``m`` as fixed time point to shutdown

        Returns
        -------
        int
            Seconds left until shutdown.
        """
        h = max(0, h)  # Force >= 0
        m = max(0, m)
        now = datetime.now()
        if aggregate:
            delta = timedelta(hours=h, minutes=m)
            out = delta.seconds
        else:
            new_time = datetime.combine(now.date(), time(hour=h, minute=m))
            diff = new_time - now
            out = diff.seconds
        return out


class ShutdownEngineWin(ShutdownEngine):
    """ShutDownizer - Windows"""

    @override
    @versionchanged("5.0.0", "Scheduled shutdown")
    def shutdown(
        self,
        h: Annotated[int, "positive"] = 0,
        m: Annotated[int, "positive"] = 0,
        aggregate: bool = True,
    ) -> None:
        time_until_sd = self._calculate_time(h=h, m=m, aggregate=aggregate)
        cmds = [f"shutdown -f -s -t {time_until_sd}"]
        self._execute_multiple_cmds(cmds)

    @override
    def restart(self, *args, **kwargs) -> None:
        cmds = ["shutdown -r"]
        self._execute_multiple_cmds(cmds)

    @override
    def sleep(self, *args, **kwargs) -> None:
        cmds = ["rundll32.exe powrprof.dll,SetSuspendState 0,1,0"]
        self._execute_multiple_cmds(cmds)

    @override
    def abort(self) -> None:
        cmds = ["shutdown -a"]
        self._execute_multiple_cmds(cmds)

    def _punish(self, *, are_you_sure_about_this: bool = False) -> None:
        """Create a `batch` script that shut down computer when boot up"""
        if not are_you_sure_about_this:
            return None
        try:
            startup_folder_win = Path(os.getenv("appdata")).joinpath(  # type: ignore
                "Microsoft", "Windows", "Start Menu", "Programs", "Startup"
            )
            with open(startup_folder_win.joinpath("system.bat"), "w") as f:
                f.write("shutdown -f -s -t 0")
        except Exception:
            logger.error("Cannot write file to startup folder")


class ShutdownEngineMac(ShutdownEngine):
    """ShutDownizer - MacOS"""

    @override
    def shutdown(self, *args, **kwargs) -> None:
        cmds = [
            ["osascript", "-e", 'tell application "System Events" to shut down'],
            "pmset sleepnow",
            "shutdown -h now",
            "sudo shutdown -h now",
        ]
        self._execute_multiple_cmds(cmds)

    @override
    def restart(self, *args, **kwargs) -> None:
        cmds = [
            ["osascript", "-e", 'tell application "System Events" to restart'],
            "shutdown -r now",
            "sudo shutdown -r now",
        ]
        self._execute_multiple_cmds(cmds)

    @override
    def sleep(self, *args, **kwargs) -> None:
        cmds = [
            ["osascript", "-e", 'tell application "System Events" to sleep'],
            "pmset sleepnow",
            "shutdown -s now",
            "sudo shutdown -s now",
        ]
        self._execute_multiple_cmds(cmds)

    @override
    def abort(self) -> None:
        cmds = [
            ["osascript", "-e", 'tell application "System Events" to cancel shutdown'],
            "killall shutdown",
            "shutdown -c",
            "sudo shutdown -c",
        ]
        self._execute_multiple_cmds(cmds)


class ShutdownEngineLinux(ShutdownEngine):
    """ShutDownizer - Linux"""

    @override
    def shutdown(self, *args, **kwargs) -> None:
        cmds = [
            "gnome-session-quit --power-off",
            "systemctl --user poweroff",
            "sudo shutdown -h now",
        ]
        self._execute_multiple_cmds(cmds)

    @override
    def restart(self, *args, **kwargs) -> None:
        cmds = [
            "gnome-session-quit --reboot",
            "systemctl reboot",
            "sudo shutdown -r now",
        ]
        self._execute_multiple_cmds(cmds)

    @override
    def sleep(self, *args, **kwargs) -> None:
        cmds = ["systemctl suspend", "sudo shutdown -s now"]
        self._execute_multiple_cmds(cmds)

    @override
    def abort(self) -> None:
        cmds = ["sudo shutdown -c"]
        self._execute_multiple_cmds(cmds)



"""
Test: Tools - Converter

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

import base64

import pytest

from absfuyu.tools.converter import Base64EncodeDecode, Text2Chemistry
from absfuyu.tools.generator import Charset, Generator


@pytest.mark.abs_tools
class TestBase64:
    """absfuyu.tools.converter.Base64EncodeDecode"""

    def test_base64_encode(self) -> None:
        test = Base64EncodeDecode.encode("Hello, World!")
        assert test == "SGVsbG8sIFdvcmxkIQ=="

    def test_base64_decode(self) -> None:
        test = Base64EncodeDecode.decode("SGVsbG8sIFdvcmxkIQ==")
        assert test == "Hello, World!"

    def test_base64_multiple(self) -> None:
        """Run multiple times"""
        TIMES = 100
        test = []
        for x in Generator.generate_string(Charset.FULL, times=TIMES):
            encode = Base64EncodeDecode.encode(x)
            test.append(x == Base64EncodeDecode.decode(encode))
        assert all(test)


@pytest.mark.abs_tools
class TestChemistryConvert:
    """absfuyu.tools.converter.Text2Chemistry"""

    @pytest.mark.parametrize(["value", "output"], [("jump", []), ("queen", [])])
    def test_convert_not_work(self, value: str, output: list) -> None:
        instance = Text2Chemistry()
        assert instance.convert(value) == output

    def test_convert(self) -> None:
        instance = Text2Chemistry()
        assert instance.convert("bakery") != []



"""
Absfuyu: Logger
---------------
Custom Logger Module

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)

Usage:
------
>>> from absfuyu.logger import logger, LogLevel
>>> logger.setLevel(LogLevel.DEBUG)
>>> logger.debug("This logs!")
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    # logger
    "logger",
    "compress_for_log",
    # log level
    "LogLevel",
]


# Library
# ---------------------------------------------------------------------------
import logging
import math
from logging.handlers import RotatingFileHandler as _RFH
from logging.handlers import TimedRotatingFileHandler as _TRFH
from pathlib import Path
from typing import Any


# Setup
# ---------------------------------------------------------------------------
class LogLevel:
    """
    ``logging``'s log level wrapper + custom log level
    """

    TRACE: int = logging.DEBUG - 5
    DEBUG: int = logging.DEBUG
    INFO: int = logging.INFO
    WARNING: int = logging.WARNING
    ERROR: int = logging.ERROR
    CRITICAL: int = logging.CRITICAL
    EXTREME: int = logging.CRITICAL + 10


class _LogFormat:
    """Some log format styles"""

    FULL = "[%(asctime)s] [%(process)-d] [%(module)s] [%(name)s] [%(funcName)s] [%(levelname)-s] %(message)s"  # Time|ProcessID|Module|Name|Function|LogType|Message
    SHORT = (
        "[%(module)s] [%(name)s] [%(funcName)s] [%(levelname)-s] %(message)s"  # Module|Name|Function|LogType|Message
    )
    CONSOLE = (
        "%(asctime)s [%(levelname)5s] %(funcName)s:%(lineno)3d: %(message)s"  # Time|LogType|Function|LineNumber|Message
    )
    FILE = "%(asctime)s [%(levelname)5s] %(filename)s:%(funcName)s:%(lineno)3d: %(message)s"  # Time|LogType|FileName|Function|LineNumber|Message


# Create a custom logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)

# Create handlers
## Console log handler
console_handler = logging.StreamHandler()
console_handler.setLevel(LogLevel.TRACE)  # Minimum log level
console_handler.setFormatter(logging.Formatter(_LogFormat.CONSOLE, datefmt="%Y-%m-%d %H:%M:%S"))
# logger.addHandler(console_handler)
logger.addHandler(logging.NullHandler())


# Functions
# ---------------------------------------------------------------------------
def _compress_list_for_print(iterable: list, max_visible: int | None = 5) -> str:
    """
    Compress the list to be more log-readable

    iterable: list
    max_visible: Maximum items can be printed on screen (Minimum: 3)
    """

    if max_visible is None or max_visible <= 2:
        max_visible = 5

    if len(iterable) <= max_visible:
        return str(iterable)
    else:
        # logger.debug(f"Max vis: {max_visible}")
        if max_visible % 2 == 0:
            cut_idx_1 = math.floor(max_visible / 2) - 1
            cut_idx_2 = math.floor(max_visible / 2)
        else:
            cut_idx_1 = cut_idx_2 = math.floor(max_visible / 2)

        # logger.debug(f"Cut pos: {(cut_idx_1, cut_idx_2)}")
        # temp = [iterable[:cut_idx_1], ["..."], iterable[len(iterable)-cut_idx_2:]]
        # out = list(chain.from_iterable(temp))
        # out = [*iterable[:cut_idx_1], "...", *iterable[len(iterable)-cut_idx_2:]] # Version 2
        out = f"{str(iterable[:cut_idx_1])[:-1]}, ..., {str(iterable[len(iterable) - cut_idx_2 :])[1:]}"  # Version 3
        # logger.debug(out)
        return f"{out} [Len: {len(iterable)}]"


def _compress_string_for_print(text: str, max_visible: int | None = 120) -> str:
    """
    Compress the string to be more log-readable

    text: str
    max_visible: Maximum text can be printed on screen (Minimum: 5)
    """

    if max_visible is None or max_visible <= 5:
        max_visible = 120

    text = text.replace("\n", " ")  # Remove new line
    # logger.debug(text)

    if len(text) <= max_visible:
        return str(text)
    else:
        cut_idx = math.floor((max_visible - 3) / 2)
        temp = f"{text[:cut_idx]}...{text[len(text) - cut_idx :]}"
        return f"{temp} [Len: {len(text)}]"


def compress_for_log(object_: Any, max_visible: int | None = None) -> str:
    """
    Compress the object to be more log-readable

    :param object_: Object
    :param max_visible: Maximum objects can be printed on screen
    :returns: Compressed log output
    :rtype: str
    """

    if isinstance(object_, list):
        return _compress_list_for_print(object_, max_visible)

    elif isinstance(object_, (set, tuple)):
        return _compress_list_for_print(list(object_), max_visible)

    elif isinstance(object_, dict):
        temp = [{k: v} for k, v in object_.items()]
        return _compress_list_for_print(temp, max_visible)

    elif isinstance(object_, str):
        return _compress_string_for_print(object_, max_visible)

    else:
        try:
            return _compress_string_for_print(str(object_), max_visible)
        except Exception:
            return object_  # type: ignore


# Class
# ---------------------------------------------------------------------------
class _CustomLogger:
    """
    Custom logger [W.I.P]

    Create a custom logger
    *Useable but maybe unstable*
    """

    def __init__(
        self,
        name: str,
        cwd: str | Path = ".",
        log_format: str | None = None,
        *,
        save_log_file: bool = False,
        separated_error_file: bool = False,
        timed_log: bool = False,
        date_log_format: str | None = None,
        error_log_size: int = 1_000_000,  # 1 MB
    ) -> None:
        """
        :param name: Custom logger name
        :param cwd: Current working directory
        :param log_format: Log format
        :param save_log_file: Save logs to log file (default: False)
        :param separated_error_file: Save error logs into a separated file (Default: False)
        :param timed_log: Split log file every day. Requirement: `save_log_file = True` (Default: False)
        :param date_log_format: Date format in log
        :param error_log_size: Error log file max size (Default: 1 MB)
        """
        self._cwd = Path(cwd)
        self.log_folder = self._cwd.joinpath("logs")
        self.log_folder.mkdir(exist_ok=True, parents=True)  # Does not throw exception when folder existed
        self.name = name
        self.log_file = self.log_folder.joinpath(f"{name}.log")

        # Create a custom logger
        try:
            self.logger = logging.getLogger(self.name)
        except Exception:
            try:
                self.logger = logging.getLogger(__name__)
            except Exception:
                self.logger = logging.getLogger()
        self.logger.setLevel(logging.DEBUG)

        if date_log_format is None:
            _date_format = "%Y-%m-%d %H:%M:%S"
        else:
            _date_format = date_log_format

        ## Console log handler
        if log_format is None:
            # Time|LogType|Function|LineNumber|Message
            _log_format = "%(asctime)s [%(levelname)5s] %(funcName)s:%(lineno)3d: %(message)s"
        else:
            _log_format = log_format
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.DEBUG)  # Minimum log level
        _console_log_format = _log_format  # Create formatters and add it to handlers
        _console_formatter = logging.Formatter(_console_log_format, datefmt=_date_format)
        console_handler.setFormatter(_console_formatter)
        self._console_handler = console_handler
        self.logger.addHandler(self._console_handler)  # Add handlers to the logger

        ## Log file handler
        if save_log_file:
            if log_format is None:
                # Time|LogType|FileName|Function|LineNumber|Message
                _log_format = "%(asctime)s [%(levelname)5s] %(filename)s:%(funcName)s:%(lineno)3d: %(message)s"
            else:
                _log_format = log_format
            file_handler = logging.FileHandler(self.log_file, mode="a", encoding="utf-8")
            file_handler.setLevel(logging.DEBUG)
            _file_log_format = _log_format
            _file_formatter = logging.Formatter(_file_log_format, datefmt=_date_format)
            file_handler.setFormatter(_file_formatter)
            self._file_handler = file_handler
            self.logger.addHandler(self._file_handler)

            if timed_log:
                ## Time handler (split log every day)
                time_handler = _TRFH(
                    self.log_folder.joinpath(f"{self.name}_timed.log"),
                    when="midnight",
                    interval=1,
                    encoding="utf-8",
                )
                time_handler.setLevel(logging.DEBUG)
                time_handler.setFormatter(_file_formatter)
                self._time_handler = time_handler
                self.logger.addHandler(self._time_handler)
                # |   Value  |    Type of interval   |
                # |:--------:|:---------------------:|
                # |     S    |        Seconds        |
                # |     M    |        Minutes        |
                # |     H    |         Hours         |
                # |     D    |          Days         |
                # |     W    |  Week day (0=Monday)  |
                # | midnight | Roll over at midnight |

        ## Error and above log handler
        if separated_error_file:
            if log_format is None:
                # Time|LogType|FileName|Function|LineNumber|Message
                _log_format = "%(asctime)s [%(levelname)5s] %(filename)s:%(funcName)s:%(lineno)3d: %(message)s"
            else:
                _log_format = log_format
            error_handler = _RFH(
                self.log_folder.joinpath(f"{self.name}_error.log"),
                maxBytes=error_log_size,
                backupCount=1,
                encoding="utf-8",
            )
            error_handler.setLevel(logging.ERROR)
            error_handler.setFormatter(_log_format)  # type: ignore
            self._error_handler = error_handler
            self.logger.addHandler(self._error_handler)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.name})"

    def __repr__(self) -> str:
        return self.__str__()

    @staticmethod
    def _add_logging_level(level_name: str, level_num: int, method_name: str | None = None):
        """
        Comprehensively adds a new logging level to the `logging` module and the
        currently configured logging class.

        `level_name` becomes an attribute of the `logging` module with the value
        `level_num`. `method_name` becomes a convenience method for both `logging`
        itself and the class returned by `logging.getLoggerClass()` (usually just
        `logging.Logger`). If `method_name` is not specified, `level_name.lower()` is
        used.

        To avoid accidental clobberings of existing attributes, this method will
        raise an `AttributeError` if the level name is already an attribute of the
        `logging` module or if the method name is already present
        """
        # Original code: https://stackoverflow.com/a/35804945/1691778
        if not method_name:
            method_name = level_name.lower()

        if hasattr(logging, level_name):
            raise AttributeError(f"{level_name} already defined in logging module")
        if hasattr(logging, method_name):
            raise AttributeError(f"{method_name} already defined in logging module")
        if hasattr(logging.getLoggerClass(), method_name):
            raise AttributeError(f"{method_name} already defined in logger class")

        # This method was inspired by the answers to Stack Overflow post
        # http://stackoverflow.com/q/2183233/2988730, especially
        # http://stackoverflow.com/a/13638084/2988730
        def logForLevel(self, message, *args, **kwargs):
            if self.isEnabledFor(level_num):
                self._log(level_num, message, args, **kwargs)

        def logToRoot(message, *args, **kwargs):
            logging.log(level_num, message, *args, **kwargs)

        logging.addLevelName(level_num, level_name)
        setattr(logging, level_name, level_num)
        setattr(logging.getLoggerClass(), method_name, logForLevel)
        setattr(logging, method_name, logToRoot)

    def add_log_level(self, level_name: str, level_num: int):
        __class__._add_logging_level(level_name, level_num)  # type: ignore
        if level_num < logging.DEBUG:
            self._console_handler.setLevel(level_num)
            self.logger.setLevel(level_num)



"""
Absfuyu: Keygen
---------------
Mod7 product key generator (90's)

This is for educational and informative purposes only.

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["Keygen"]


# Library
# ---------------------------------------------------------------------------
import random

from absfuyu.core import BaseClass


# Class
# ---------------------------------------------------------------------------
class Keygen(BaseClass):
    """Key generator"""

    @staticmethod
    def _is_mod7(text: str) -> bool:
        """
        Check if sum of elements in a string is divisible by 7

        :param text: number in str type to check
        """
        text = str(text)  # Safety convert
        try:
            res = sum(map(int, text)) % 7
            # logger.debug(f"Sum value: {res}")
            return res == 0
        except Exception:
            raise ValueError("Invalid string")  # noqa: B904

    @classmethod
    def _mod7_gen(cls, num_of_digits: int, *, fillchar: str = "0") -> str:
        """
        Generate a number with desired length that is divisible by 7

        Parameters
        ----------
        num_of_digits : int
            Length of number

        fillchar : str
            filled character when `len(generated number)` < `num_of_digits`

        Returns
        -------
        str:
            Mod7 number
        """

        # Init
        mod7_num: int = 0

        # Conditions
        max_value = 10**num_of_digits - 1
        mod7_valid = False
        invalid_digits = [0, 8, 9]  # Invalid last digit

        # Loop
        while not mod7_valid:
            # Gen num
            mod7_num = random.randint(0, max_value)

            # Check last digit
            if int(str(mod7_num)[-1]) in invalid_digits:
                continue

            # Check divide by 7
            if cls._is_mod7(str(mod7_num)):
                mod7_valid = True

        # Output
        return str(mod7_num).rjust(num_of_digits, fillchar)

    @classmethod
    def mod7_cd_key(cls, fast: bool = False) -> str:
        """
        CD Key generator

        Format: ``XXX-XXXXXXX``

        Rules:

        - Last seven digits must add to be divisible by ``7``
        - First 3 digits cannot be ``333``, ``444``,..., ``999``
        - Last digit of last seven digits cannot be ``0``, ``8`` or ``9``

        Parameters
        ----------
        fast : bool
            Use pre-generated key
            (Default: ``False``)

        Returns
        -------
        str:
            Mod7 Key
        """

        # Fast mode: pre-generated key
        if fast:
            return "111-1111111"

        # PART 01
        part1_valid = False
        part1_not_valid_digits = [333, 444, 555, 666, 777, 888]
        part1: str = ""
        while not part1_valid:  # Loop check
            part1_num = random.randint(0, 998)  # Gen random int from 0 to 998
            # part1_num = random.randint(100,300) # or just use this
            if part1_num not in part1_not_valid_digits:
                part1 = str(part1_num).rjust(3, "0")  # Convert into string
                part1_valid = True  # Break loop

        # PART 02
        part2 = cls._mod7_gen(num_of_digits=7)

        # OUTPUT
        return f"{part1}-{part2}"

    @classmethod
    def mod7_11_digit_key(cls, fast: bool = False) -> str:
        """
        11-digit CD Key generator

        Format: ``XXXX-XXXXXXX``

        - ``XXXX``: Can be anything from ``0001`` to ``9991``. The last digit must be 3rd digit + ``1`` or ``2``. When the result is > ``9``, it overflows to ``0`` or ``1``.
        - ``XXXXXXX``: Same as CD Key

        Parameters
        ----------
        fast : bool
            Use pre-generated key
            (Default: ``False``)

        Returns
        -------
        str:
            Mod7 Key
        """

        # Fast mode: pre-generated key
        if fast:
            return "0001-0000007"

        # PART 01
        part1_valid = False
        part1: str = ""
        while not part1_valid:
            part1_1_num = random.randint(0, 999)  # Random 3-digit number
            last_digit_choice = [1, 2]  # Choice for last digit
            part1_2_num = int(str(part1_1_num)[-1]) + random.choice(
                last_digit_choice
            )  # Make last digit
            if part1_2_num > 9:  # Check condition then overflow
                part1_2_num = int(str(part1_2_num)[-1])
            part1_str = f"{part1_1_num}{part1_2_num}"  # Concat string
            if int(part1_str) < 9991:  # Check if < 9991
                part1 = part1_str.rjust(4, "0")
                part1_valid = True

        # PART 02
        part2 = cls._mod7_gen(num_of_digits=7)

        # OUTPUT
        return f"{part1}-{part2}"

    @classmethod
    def mod7_oem_key(cls, fast: bool = False) -> str:
        """
        OEM Key generator

        Format: ``ABCYY-OEM-0XXXXXX-XXXXX``

        - ``ABC``: The day of the year. It can be any value from ``001`` to ``366``
        - ``YY``: The last two digits of the year. It can be anything from ``95`` to ``03``
        - ``0XXXXXX``: A random number that has a sum that is divisible by ``7`` and does not end with ``0``, ``8`` or ``3``.
        - ``XXXXX``: A random 5-digit number

        Parameters
        ----------
        fast : bool
            Use pre-generated key
            (Default: ``False``)

        Returns
        -------
        str:
            Mod7 Key
        """

        # Fast mode: pre-generated key
        if fast:
            return "00100-OEM-0000007-00000"

        # PART ABC
        abc_part = str(random.randint(1, 365)).rjust(3, "0")

        # PART YY
        year_choice = ["95", "96", "97", "98", "99", "00", "01", "02"]
        # isleap = [False, True, False, False, False, True, False, False]
        # year_choice = ["95", "96", "97", "98", "99", "00", "01", "02", "03"] # "03" not wotk on win95
        y_part = random.choice(year_choice)

        # NUM PART
        num_part = cls._mod7_gen(num_of_digits=6)

        # NUM PART 02
        num_part_2 = str(random.randint(0, 99999)).rjust(5, "0")

        # OUTPUT
        return f"{abc_part}{y_part}-OEM-0{num_part}-{num_part_2}"

    @classmethod
    def mod7_combo(cls, fast: bool = False):
        """
        A combo that consist of CD, 11-digit, and OEM Key

        Parameters
        ----------
        fast : bool
            Use pre-generated key
            (Default: ``False``)

        Returns
        -------
        dict:
            Mod7 Key combo
        """
        out = {
            "CD Key": cls.mod7_cd_key(fast=fast),
            "OEM Key": cls.mod7_oem_key(fast=fast),
            "11-digit Key": cls.mod7_11_digit_key(fast=fast),
        }
        return out



"""
Absfuyu: Content
----------------
Handle .txt file

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)

Usage:
------
>>> from absfuyu.general.content import ContentLoader
"""

# TODO: rewrite this

# Module level
# ---------------------------------------------------------------------------
__all__ = ["ContentLoader", "Content", "LoadedContent"]


# Library
# ---------------------------------------------------------------------------
import json
import random
import re
from collections import Counter
from itertools import chain
from pathlib import Path
from typing import Self

from absfuyu.core import BaseClass, GetClassMembersMixin, unidecode
from absfuyu.logger import logger


# Class
# ---------------------------------------------------------------------------
class Content(BaseClass):
    """
    Contain data

    Data format: list[str, list[str]]
        where: ``str: data``; ``list[str]: data tags``
    """

    def __init__(self, data: list) -> None:
        self.data: str = str(data[0])
        self.tag: list = data[1]
        # logger.debug(self.__dict__)

    def __str__(self) -> str:
        # return f"{self.data} | {self.tag}"
        return str(self.data)

    def unidecoded(self) -> Self:
        """
        Convert data through ``unidecode`` package

        Returns
        -------
        Content
            ``unidecoded`` Content
        """
        return self.__class__([unidecode(self.data), list(map(unidecode, self.tag))])

    def to_text(self) -> str:
        """
        Convert back into text

        Returns
        -------
        str
            Text
        """
        # hard code
        tags = ",".join(self.tag)
        return f"{self.data}|{tags}"

    def short_form(self, separator: str = ",") -> str:
        """
        Short form show only first item when separated by ``separator``

        Parameters
        ----------
        separator : str
            Default: ``","``

        Returns
        -------
        str
            Short form
        """
        if not separator.startswith(","):
            logger.debug(f"Separated symbol: {separator}")
        return self.data.split(separator)[0]

    def handle_address(
        self, address_separator: str = ",", *, first_item_not_address: bool = True
    ) -> Self:
        """
        Handle ``self.data`` as address and then update the address into ``self.tag``

        Parameters
        ----------
        address_separator : str
            | Split the address by which character
            | (Default: ``","``)

        first_item_not_address : bool
            Set to ``False`` when ``<splited data>[0]`` is not part of an address

        Returns
        -------
        Content
            Handled Content


        Example:
        --------
        >>> test = "Shop A, 22 ABC Street, DEF District, GHI City"
        >>> # After handle_address()
        ["Shop A", "22 ABC Street", "DEF District", "GHI City"]

        >>> # After handle_address(first_item_not_address = False)
        ["22 ABC Street", "DEF District", "GHI City"]
        """
        if first_item_not_address:
            temp = self.data.split(address_separator)
        else:
            logger.debug(
                f"First item ({self.data.split(address_separator)[0]}) is not part of an address"
            )
            temp = self.data.split(address_separator)[1:]

        new_tag = [x.strip().lower() for x in temp]
        logger.debug(f"Current tags: {self.tag}")
        logger.debug(f"New tags: {new_tag}")
        self.tag = list(set(self.tag + new_tag))
        logger.debug(f"Final tags: {self.tag} Len: {len(self.tag)}")

        return self.__class__([self.data, self.tag])


class LoadedContent(list[Content], GetClassMembersMixin):
    """
    Contain list of ``Content``
    """

    def __str__(self) -> str:
        # return f"{self.__class__.__name__} - Total: {len(self)}"
        return f"{self.__class__.__name__}({[x.data for x in self]})"

    def __repr__(self) -> str:
        return self.__str__()

    def apply(self, func) -> Self:
        """
        Apply function to each entry

        :param func: Callable function
        :type func: Callable
        :rtype: LoadedContent
        """
        return self.__class__(func(x.data) for x in self)

    @classmethod
    def load_from_json(cls, file: Path) -> Self:
        """
        Use this method to load data from ``.json`` file from ``to_json()`` method

        Parameters
        ----------
        file : Path
            Path to ``.json`` file

        Returns
        -------
        LoadedContent
            Loaded content from ``.json`` file
        """
        with open(file) as json_file:
            parsed_json: dict = json.load(json_file)
        out = [Content(list(x.values())) for x in parsed_json]
        return cls(out)

    @property
    def tags(self) -> list:
        """A list contains all available tag"""
        temp = chain.from_iterable([x.tag for x in self])
        out = list(set(temp))
        logger.debug(f"Found {len(out)} {'tags' if len(out) > 1 else 'tag'}")
        return sorted(out)

    def tag_count(self) -> Counter:
        """
        Count number of tags

        :rtype: Counter
        """
        temp = chain.from_iterable([x.tag for x in self])
        logger.debug(temp)
        return Counter(temp)

    def filter(self, tag: str) -> Self:
        """
        Filter out entry with ``tag``

        :param tag: Tag to filter
        :type tag: str
        :rtype: LoadedContent
        """
        tag = tag.strip().lower()
        logger.debug(f"Tag: {tag}")
        if tag not in self.tags:
            # tag = random.choice(self.tags)
            # logger.debug(f"Tag not exist, changing to a random tag... {tag}")
            logger.warning(f'"{tag}" tag does not exist')
            _avail_tag = ", ".join(list(dict(self.tag_count().most_common(5)).keys()))
            raise ValueError(
                f"Available tags: {_avail_tag},..."
                f"\nMore tag at: `{self.__class__.__name__}.tags`"
            )
        return self.__class__([x for x in self if tag in x.tag])

    def find(self, keyword: str) -> Self:
        """
        Return all entries that include ``keyword``

        :param keyword: Keyword to find
        :type keyword: str
        :rtype: LoadedContent
        """
        temp = self.__class__(
            [x for x in self if x.data.lower().find(keyword.lower()) >= 0]
        )
        if temp:
            logger.debug(f"Found {len(temp)} {'entries' if len(temp) > 1 else 'entry'}")
        else:
            logger.debug("No result")
        return temp

    def short_form(self, separator: str = ",") -> list[str]:
        """
        Shows only first item when separated by ``separator`` of ``Content.data``

        Parameters
        ----------
        separator : str
            Default: ``","``

        Returns
        -------
        list[str]
            Short form
        """
        return [x.short_form(separator) for x in self]

    def pick_one(self, tag: str | None = None) -> Content:
        """
        Pick a random entry

        Parameters
        ----------
        tag : str | None
            Pick a random entry with ``tag``
            (Default: None)

        Returns
        -------
        Content
            Random entry
        """
        if tag:
            temp = self.filter(tag)
            logger.debug(f"Tag: {tag}")
            return random.choice(temp)
        return random.choice(self)

    def handle_address(
        self, address_separator: str = ",", *, first_item_not_address: bool = True
    ) -> Self:
        """
        Execute ``Content.handle_address()`` on every ``self.data`` of ``Content``

        Parameters
        ----------
        address_separator : str
            | Split the address by which character
            | (Default: ``","``)

        first_item_not_address : bool
            Set to ``False`` when ``<splited data>[0]`` is not part of an address

        Returns
        -------
        LoadedContent
            Handled Content


        Example:
        --------
        >>> test = "Shop A, 22 ABC Street, DEF District, GHI City"
        >>> # After handle_address()
        ["Shop A", "22 ABC Street", "DEF District", "GHI City"]

        >>> # After handle_address(first_item_not_address = False)
        ["22 ABC Street", "DEF District", "GHI City"]
        """
        return self.__class__(
            [
                x.handle_address(
                    address_separator=address_separator,
                    first_item_not_address=first_item_not_address,
                )
                for x in self
            ]
        )

    def to_json(self, no_accent: bool = False) -> str:
        """
        Convert data into ``.json`` file

        Parameters
        ----------
        no_accent : bool
            | when ``True``: convert the data through ``unidecode`` package
            | (Default: ``False``)

        Returns
        -------
        str
            Data
        """
        if no_accent:
            out = [x.unidecoded().__dict__ for x in self]
        else:
            out = [x.__dict__ for x in self]
        logger.debug(out)
        return json.dumps(out, indent=2)

    def to_text(self, no_accent: bool = False) -> str:
        """
        Convert data into ``.txt`` file

        Parameters
        ----------
        no_accent : bool
            | when ``True``: convert the data through ``unidecode`` package
            | (Default: ``False``)

        Returns
        -------
        str
            Data
        """
        if no_accent:
            out = [x.unidecoded().to_text() for x in self]
        else:
            out = [x.to_text() for x in self]
        logger.debug(out)
        return "\n".join(out)


class ContentLoader(BaseClass):
    """
    This load data from ``.txt`` file

    Content format:
    ``<content><split_symbol><tags separated by <tag_separate_symbol>>``
    """

    def __init__(
        self,
        file_path: Path | str,
        characteristic_detect: bool = True,
        tag_dictionary: dict | None = None,
        *,
        comment_symbol: str = "#",
        split_symbol: str = "|",
        tag_separate_symbol: str = ",",
    ) -> None:
        """
        Parameters
        ----------
        file_path : str
            file location/file to load

        characteristic_detect : bool
            detect whether the content is long, short, or a question
            (default: ``True``)

        tag_dict : dict
            custom tag pattern
            format: ``{"keyword": "tag",...}``
            example: ``{"apple": "fruit", "orange": "fruit"}``

        comment_symbol : str
            symbol that `ContentLoader` will ignore
            (default: ``"#"``)

        split_symbol : str
            symbol that `ContentLoader` will split content and tags
            (default: ``"|"``)

        tag_separate_symbol : str
            symbol that `ContentLoader` will split between tags
            (default: ``","``)
        """
        # file
        self.file = Path(file_path)

        # characteristic detect
        self.characteristic_detect: bool = characteristic_detect

        # tag dictionary
        if tag_dictionary is None:
            # logger.debug("No tag patern available")
            self.tag_dictionary: dict = dict()
        else:
            logger.debug(
                f"Tag pattern available: "
                f"{len(tag_dictionary)} {'entries' if len(tag_dictionary) > 1 else 'entry'} "
                f"({len(set(tag_dictionary.values()))} unique "
                f"{'tags' if len(set(tag_dictionary.values())) > 1 else 'tag'})"
            )
            self.tag_dictionary = tag_dictionary

        # symbol stuff
        assert comment_symbol != split_symbol, (
            "comment_symbol and split_symbol should have different values"
        )
        assert tag_separate_symbol != split_symbol, (
            "tag_separate_symbol and split_symbol should have different values"
        )
        self.comment_symbol: str = comment_symbol
        self.split_symbol: str = split_symbol
        self.tag_separate_symbol: str = tag_separate_symbol

    def content_format(self) -> str:
        """
        Shows current content format

        Returns
        -------
        str
            Current content format


        Example:
        --------
        >>> ContentLoader(".").content_format()
        # This line will be ignored
        <content> | <tag1>, <tag2>
        """
        out = (
            f"{self.comment_symbol} This line will be ignored\n"
            f"<content> {self.split_symbol} "
            f"<tag1>{self.tag_separate_symbol} <tag2>, ..., <tag n>"
        )
        return out

    def load(self) -> list:
        """
        Load content from ``self.file``

        Returns
        -------
        list
            List of content
        """
        with open(self.file, "r", encoding="utf-8") as data:
            logger.debug("Loading data...")
            dat = []
            check = []

            for i, x in enumerate(data.readlines()):
                x = x.strip()
                if x.startswith(self.comment_symbol) or len(x) == 0:
                    continue  # skip comment and empty lines
                logger.debug(
                    f"### Loop {i + 1} #####################################################################"
                )

                temp = x.split(self.split_symbol)
                if len(temp) != 2:
                    logger.debug(f"Split len: {len(temp)}")
                    logger.warning(
                        f"The current entry is missing data or tag: {x[:20]}..."
                    )

                temp[0] = temp[0].strip()
                if temp[0].lower() not in check:
                    check.append(temp[0].lower())  # check for dupes

                    # tag
                    additional_tags = []
                    for k, v in self.tag_dictionary.items():
                        key = k.strip().lower()
                        val = temp[0].lower()
                        regex_pattern = f"[^a-zA-Z0-9]({key})[^a-zA-Z0-9]|^({key})[^a-zA-Z0-9]|[^a-zA-Z0-9]({key})$"
                        if re.search(regex_pattern, val) is not None or val.startswith(
                            key
                        ):
                            # regex has a bug (or idk if it's a bug or not) that
                            # doesn't recognise when there is only one word in the sentence
                            # therefore use `val.startswith(key)` to fix
                            additional_tags.append(v.strip().lower())

                    if self.characteristic_detect:
                        long_short = (120, 20)  # setting
                        if len(temp[0]) > long_short[0]:
                            additional_tags.append("long")
                        if len(temp[0]) < long_short[1]:
                            additional_tags.append("short")
                        if temp[0][-1].startswith("?"):
                            additional_tags.append("question")
                    if additional_tags:
                        logger.debug(f"Additional tags: {additional_tags}")

                    try:
                        tags = [
                            tag.strip() for tag in temp[1].strip().lower().split(",")
                        ]  # separate and strip tags
                        logger.debug(f"Tags: {tags}")
                    except Exception:
                        logger.warning("No tag found in the original string")
                        if additional_tags:
                            tags = additional_tags
                        else:
                            tags = ["unspecified"]
                            logger.debug('Assigned "unspecified" tag')

                    tags.extend(additional_tags)
                    final_tags = list(set(tags))
                    logger.debug(f"Final tags: {final_tags} Len: {len(final_tags)}")

                    dat.append([temp[0], final_tags])  # add everything
                else:
                    logger.debug(f"Found duplicates, {x} removed")

        return dat

    def load_content(self) -> LoadedContent:
        """
        Load data into a list of ``Content``

        Returns
        -------
        LoadedContent
            Loaded content
        """
        return LoadedContent(map(Content, self.load()))



"""
Test: Passwordlib

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

from itertools import combinations_with_replacement

import pytest

from absfuyu.dxt import Text
from absfuyu.tools.passwordlib import TOTP, PasswordGenerator

# def test_generate_password():
#     test = [password_check(Password.generate_password()) for _ in range(100)]
#     assert all(test)


class TestPasswordlib:
    def test_generate_password_matrix(self) -> None:
        num_of_test = 1000

        def check(value: dict) -> int:
            return sum([1 for x in value.values() if x > 0])

        out = []
        check_matrix = list(
            set(
                combinations_with_replacement(
                    [True, False, True, False, True, False], 3
                )
            )
        )
        for x in check_matrix:
            include_number, include_special, include_uppercase = x
            check_value = sum(x) + 1

            test = [
                Text(
                    PasswordGenerator.generate_password(
                        include_number=include_number,
                        include_special=include_special,
                        include_uppercase=include_uppercase,
                    )
                ).analyze()
                for _ in range(num_of_test)
            ]
            test = list(set(map(check, test)))
            if len(test) == 1:
                out.append(test[0] == check_value)
            else:
                # assert False
                raise AssertionError()

        assert all(out)

    def test_generate_passphrase(self) -> None:
        _ = PasswordGenerator.generate_passphrase(
            num_of_blocks=3,
            block_divider="-",
            first_letter_cap=True,
            include_number=True,
        )


class TestTOTP:
    def test_totp_initialization(self):
        totp = TOTP(secret="BASE32SECRET")
        assert totp.secret == "BASE32SECRET"
        assert totp.name == "None"
        assert totp.issuer is None
        assert totp.algorithm == "SHA1"
        assert totp.digit == 6
        assert totp.period == 30

    def test_totp_initialization_with_optional_parameters(self):
        totp = TOTP(
            secret="ANOTHERSECRET",
            name="TestUser",
            issuer="TestIssuer",
            algorithm="SHA256",
            digit=8,
            period=60,
        )
        assert totp.secret == "ANOTHERSECRET"
        assert totp.name == "TestUser"
        assert totp.issuer == "TestIssuer"
        assert totp.algorithm == "SHA256"
        assert totp.digit == 8
        assert totp.period == 60

    def test_totp_secret_is_uppercase(self):
        totp = TOTP(secret="lowercase")
        assert totp.secret == "LOWERCASE"

    def test_totp_digit_and_period_are_at_least_one(self):
        totp = TOTP(secret="SECRET", digit=0, period=0)
        assert totp.digit == 1
        assert totp.period == 1

    def test_to_url_with_all_parameters(self):
        totp = TOTP(
            secret="FULLSECRET",
            name="Full User",
            issuer="FullIssuer",
            algorithm="SHA512",
            digit=8,
            period=120,
        )
        url = totp.to_url()
        expected_url = (
            "otpauth://totp/Full%20User?"
            "secret=FULLSECRET&"
            "issuer=FullIssuer&"
            "algorithm=SHA512&"
            "digit=8&"
            "period=120"
        )
        assert url == expected_url



"""
Absfuyu: Package
----------------
Package related

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["PackageManager"]


# Library
# ---------------------------------------------------------------------------
import ensurepip
import importlib
import importlib.metadata
import importlib.util
import logging
import subprocess
import sys
from typing import overload

from absfuyu.core import BaseClass


# Class
# ---------------------------------------------------------------------------
class PackageManager(BaseClass):
    """Utility class for checking, installing, and importing Python packages safely."""

    @overload
    def __init__(self, package: str, import_name: str | None = None) -> None: ...
    @overload
    def __init__(
        self, package: str, import_name: str | None = None, auto_upgrade: bool = False, auto_bootstrap: bool = True
    ) -> None: ...
    def __init__(
        self,
        package: str,
        import_name: str | None = None,
        auto_upgrade: bool = False,
        auto_bootstrap: bool = True,
        logger: logging.Logger | None = None,
    ) -> None:
        """
        A configurable package manager utility that can check, install, and import Python packages safely.

        Parameters
        ----------
        package : str
        The pip package name (e.g. "absfuyu", "random").

        import_name : str | None
            The importable module name if different from the package name (e.g. "PIL" for "pillow").

        auto_upgrade : bool, by default ``False``
            Whether to automatically upgrade packages during installation.

        auto_bootstrap : bool, by default ``True``
            Whether to automatically bootstrap pip using ensurepip if it's missing.
        """
        self.package = package
        self.import_name = import_name
        self.auto_upgrade = auto_upgrade
        self.auto_bootstrap = auto_bootstrap

        # Not available yet
        self.logger = logger

    @property
    def version(self) -> str | None:
        """
        Version of package if available

        Returns
        -------
        str
            Version of package

        None
            When package is not available
        """
        try:
            return importlib.metadata.version(self.package or self.import_name)
        except importlib.metadata.PackageNotFoundError:
            return None

    def _run_pip_install(self, package: str) -> None:
        """
        Internal helper to safely run pip install, bootstrapping pip if needed.
        """
        try:
            import pip  # noqa: F401
        except ImportError:
            if self.auto_bootstrap:
                ensurepip.bootstrap()
            else:
                raise RuntimeError("pip not found and auto_bootstrap is disabled.")

        cmd = [sys.executable, "-m", "pip", "install", package]
        if self.auto_upgrade:
            cmd.append("--upgrade")

        try:
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError as e:
            raise RuntimeError(f"Failed to install {package}: {e}")

    def ensure_installed(self) -> None:
        """
        Ensure a Python package is installed. Installs (and optionally upgrades) it if missing.
        """
        module_name = self.import_name or self.package

        if importlib.util.find_spec(module_name) is None:
            self._run_pip_install(self.package)

    def ensure_import(self):
        """
        Ensure a Python package is importable, installing it if necessary.
        Returns the imported module.
        """
        module_name = self.import_name or self.package

        if importlib.util.find_spec(module_name) is not None:
            return importlib.import_module(module_name)

        if self.logger is not None:
            self.logger.info(f"Installing missing package: {self.package} ...")

        self._run_pip_install(self.package)
        return importlib.import_module(module_name)



"""
Absfuyu: Data Extension
-----------------------
int extension

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = ["IntExt", "Pow"]


# Library
# ---------------------------------------------------------------------------
import math
from abc import ABC
from collections import Counter
from typing import Any, Literal, Self, overload

try:
    from typing import override  # type: ignore
except ImportError:
    from absfuyu.core.decorator import dummy_decorator as override

from absfuyu.core.baseclass import GetClassMembersMixin
from absfuyu.core.docstring import deprecated, versionadded, versionchanged
from absfuyu.dxt.dxt_support import DictBoolTrue


# Class
# ---------------------------------------------------------------------------
class Pow:
    """Number power by a number"""

    def __init__(self, number: int | float, power_by: int) -> None:
        self.number = number
        self.power_by = power_by

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        if self.power_by == 1:
            return str(self.number)
        else:
            return f"{self.number}^{self.power_by}"
        # return f"{self.__class__.__name__}({self.number}, {self.power_by})"

    def to_list(self) -> list[int]:
        """
        Convert into list

        Returns
        -------
        list[int | float]
            list
        """
        return [self.number] * self.power_by  # type: ignore

    def calculate(self) -> float:
        """
        Calculate the ``self.number`` to the power of ``self.power_by``

        Returns
        -------
        float
            Result
        """
        # return self.number**self.power_by
        return math.pow(self.number, self.power_by)


class IntExt(GetClassMembersMixin, int):
    """
    ``int`` extension

    >>> # For a list of new methods
    >>> IntExt.show_all_methods()
    """

    # convert
    @deprecated("5.4.0", reason="Use format(...) instead")
    def to_binary(self) -> str:
        """
        Convert to binary number

        Returns
        -------
        str
            Binary number


        Example:
        --------
        >>> test = IntExt(10)
        >>> test.to_binary()
        '1010'
        """
        return format(self, "b")

    def to_celcius_degree(self) -> float:
        """
        Convert into Celcius degree as if ``self`` is Fahrenheit degree

        Returns
        -------
        float
            Celcius degree


        Example:
        --------
        >>> test = IntExt(10)
        >>> test.to_celcius_degree()
        -12.222222222222221
        """
        c_degree = (self - 32) / 1.8
        return c_degree

    def to_fahrenheit_degree(self) -> float:
        """
        Convert into Fahrenheit degree as if ``self`` is Celcius degree

        Returns
        -------
        float
            Fahrenheit degree


        Example:
        --------
        >>> test = IntExt(10)
        >>> test.to_fahrenheit_degree()
        50.0
        """
        f_degree = (self * 1.8) + 32
        return f_degree

    def reverse(self) -> Self:
        """
        Reverse a number. Reverse ``abs(number)`` if ``number < 0``

        Returns
        -------
        Self
            Reversed number.


        Example:
        --------
        >>> test = IntExt(102)
        >>> test.reverse()
        201
        """
        number = int(self)
        if number <= 1:
            number *= -1
        return self.__class__(str(number)[::-1])

    # is_stuff
    def is_even(self) -> bool:
        """
        An even number is a number which divisible by 2

        Returns
        -------
        bool
            | ``True`` if an even number
            | ``False`` if not an even number
        """
        return self % 2 == 0

    def is_prime(self) -> bool:
        """
        Check if the integer is a prime number or not

            A prime number is a natural number greater than ``1``
            that is not a product of two smaller natural numbers.
            A natural number greater than ``1`` that is not prime
            is called a composite number.

        Returns
        -------
        bool
            | ``True`` if a prime number
            | ``False`` if not a prime number
        """
        number = self

        if number <= 1:
            return False
        for i in range(2, int(math.sqrt(number)) + 1):  # divisor range
            if number % i == 0:
                return False
        return True

    def is_twisted_prime(self) -> bool:
        """
        A number is said to be twisted prime if
        it is a prime number and
        reverse of the number is also a prime number

        Returns
        -------
        bool
            | ``True`` if a twisted prime number
            | ``False`` if not a twisted prime number
        """
        prime = self.is_prime()
        rev = self.reverse().is_prime()
        return prime and rev

    def is_perfect(self) -> bool:
        """
        Check if integer is perfect number

            Perfect number: a positive integer that is
            equal to the sum of its proper divisors.
            The smallest perfect number is ``6``, which is
            the sum of ``1``, ``2``, and ``3``.
            Other perfect numbers are ``28``, ``496``, and ``8,128``.

        Returns
        -------
        bool
            | ``True`` if a perfect number
            | ``False`` if not a perfect number
        """
        # ---
        """
        # List of known perfect number
        # Source: https://en.wikipedia.org/wiki/List_of_Mersenne_primes_and_perfect_numbers
        perfect_number_index = [
            2, 3, 5, 7,
            13, 17, 19, 31, 61, 89,
            107, 127, 521, 607,
            1279, 2203, 2281, 3217, 4253, 4423, 9689, 9941,
            11_213, 19_937, 21_701, 23_209, 44_497, 86_243,
            110_503, 132_049, 216_091, 756_839, 859_433,
            # 1_257_787, 1_398_269, 2_976_221, 3_021_377, 6_972_593,
            # 13_466_917, 20_996_011, 24_036_583, 25_964_951,
            # 30_402_457, 32_582_657, 37_156_667, 42_643_801,
            # 43_112_609, 57_885_161,
            ## 74_207_281, 77_232_917, 82_589_933
        ]
        perfect_number = []
        for x in perfect_number_index:
            # a perfect number have a form of (2**(n-1))*((2**n)-1)
            perfect_number.append((2**(x-1))*((2**x)-1))
        """
        number = int(self)

        perfect_number = [
            6,
            28,
            496,
            8128,
            33_550_336,
            8_589_869_056,
            137_438_691_328,
            2_305_843_008_139_952_128,
        ]

        if number in perfect_number:
            return True

        elif number < perfect_number[-1]:
            return False

        else:
            # Faster way to check
            perfect_number_index: list[int] = [
                61,
                89,
                107,
                127,
                521,
                607,
                1279,
                2203,
                2281,
                3217,
                4253,
                4423,
                9689,
                9941,
                11_213,
                19_937,
                21_701,
                23_209,
                44_497,
                86_243,
                110_503,
                132_049,
                216_091,
                756_839,
                859_433,
                1_257_787,
                # 1_398_269,
                # 2_976_221,
                # 3_021_377,
                # 6_972_593,
                # 13_466_917,
                # 20_996_011,
                # 24_036_583,
                # 25_964_951,
                # 30_402_457,
                # 32_582_657,
                # 37_156_667,
                # 42_643_801,
                # 43_112_609,
                # 57_885_161,
                ## 74_207_281,
                ## 77_232_917,
                ## 82_589_933
            ]
            for x in perfect_number_index:
                # a perfect number have a form of (2**(n-1))*((2**n)-1)
                perfect_number = (2 ** (x - 1)) * ((2**x) - 1)
                if number < perfect_number:  # type: ignore
                    return False
                elif number == perfect_number:  # type: ignore
                    return True

            # Manual way when above method not working
            # sum
            s = 1
            # add all divisors
            i = 2
            while i * i <= number:
                if number % i == 0:
                    s += +i + number / i  # type: ignore
                i += 1
            # s == number -> perfect
            return True if s == number and number != 1 else False

    def is_narcissistic(self) -> bool:
        """
        Check if a narcissistic number

            In number theory, a narcissistic number
            (also known as a pluperfect digital invariant (PPDI),
            an Armstrong number (after Michael F. Armstrong)
            or a plus perfect number) in a given number base ``b``
            is a number that is the sum of its own digits
            each raised to the power of the number of digits.

        Returns
        -------
        bool
            | ``True`` if a narcissistic number
            | ``False`` if not a narcissistic number
        """
        try:
            check = sum([int(x) ** len(str(self)) for x in str(self)])
            res = int(self) == check
            return res  # type: ignore
        except Exception:
            return False

    def is_palindromic(self) -> bool:
        """
        A palindromic number (also known as a numeral palindrome
        or a numeric palindrome) is a number (such as ``16461``)
        that remains the same when its digits are reversed.

        Returns
        -------
        bool
            | ``True`` if a palindromic number
            | ``False`` if not a palindromic number
        """
        return self == self.reverse()

    def is_palindromic_prime(self) -> bool:
        """
        A palindormic prime is a number which is both palindromic and prime

        Returns
        -------
        bool
            | ``True`` if a palindormic prime number
            | ``False`` if not a palindormic prime number
        """
        return self.is_palindromic() and self.is_prime()

    # calculation stuff
    def lcm(self, with_number: int) -> Self:
        """
        Least common multiple of ``self`` and ``with_number``

        Parameters
        ----------
        with_number : int
            The number that want to find LCM with

        Returns
        -------
        Self
            Least common multiple.


        Example:
        --------
        >>> test = IntExt(102)
        >>> test.lcm(5)
        510
        """
        return self.__class__(math.lcm(self, with_number))

    @versionchanged("3.3.0", reason="Updated functionality")
    def gcd(self, with_number: int) -> Self:
        """
        Greatest common divisor of ``self`` and ``with_number``

        Parameters
        ----------
        with_number : int
            The number that want to find GCD with

        Returns
        -------
        Self
            Greatest common divisor.


        Example:
        --------
        >>> test = IntExt(1024)
        >>> test.gcd(8)
        8
        """
        return self.__class__(math.gcd(self, with_number))

    def add_to_one_digit(self, master_number: bool = False) -> Self:
        """
        Convert ``self`` into 1-digit number
        by adding all of the digits together

        Parameters
        ----------
        master_number : bool
            | Break when sum = ``22`` or ``11`` (numerology)
            | (Default: ``False``)

        Returns
        -------
        Self
            IntExt


        Example:
        --------
        >>> test = IntExt(119)
        >>> test.add_to_one_digit()
        2

        >>> test = IntExt(119)
        >>> test.add_to_one_digit(master_number=True)
        11
        """

        number = int(self)
        if number < 0:  # Convert positive
            number *= -1

        while len(str(number)) != 1:
            number = sum(map(int, str(number)))
            if master_number:
                if number == 22 or number == 11:
                    break  # Master number

        return self.__class__(number)

    @versionchanged("5.0.0", reason="Removed ``short_form`` parameter")
    def divisible_list(self) -> list[int]:
        """
        A list of divisible number

        Returns
        -------
        list[int]
            A list of divisible number


        Example:
        --------
        >>> test = IntExt(1024)
        >>> test.divisible_list()
        [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]
        """

        if self <= 1:
            return [1]
        divi_list = [x for x in range(1, int(self / 2) + 1) if self % x == 0] + [self]

        return divi_list

    @overload
    def prime_factor(self) -> list[Pow]: ...  # type: ignore

    @overload
    def prime_factor(self, short_form: Literal[False] = ...) -> list[int]: ...

    def prime_factor(self, short_form: bool = True) -> list[int] | list[Pow]:
        """
        Prime factor

        Parameters
        ----------
        short_form : bool
            | Show prime list in short form
            | Normal example: ``[2, 2, 2, 3, 3]``
            | Short form example: ``[2^3, 3^2]``
            | (Default: ``True``)

        Returns
        -------
        list[int] | list[Pow]
            | List of prime number that when multiplied together == ``self``
            | list[int]: Long form
            | list[Pow]: Short form


        Example:
        --------
        >>> test = IntExt(1024)
        >>> test.prime_factor()
        [2^10]

        >>> test = IntExt(1024)
        >>> test.prime_factor(short_form=False)
        [2, 2, 2, 2, 2, 2, 2, 2, 2, 2]
        """

        # Generate list
        factors = []
        divisor = 2
        number = int(self)
        if number <= 1:
            return [number]
        while divisor <= number:
            if number % divisor == 0:
                factors.append(divisor)
                number //= divisor  # number = number // divisor
            else:
                divisor += 1

        # Output
        if short_form:
            temp = dict(Counter(factors))
            return [Pow(k, v) for k, v in temp.items()]
        return factors

    # analyze
    def analyze(self, short_form: bool = True) -> dict[str, dict[str, Any]]:
        """
        Analyze the number with almost all ``IntExt`` method

        Parameters
        ----------
        short_form : bool
            | Enable short form for some items
            | (Default: ``True``)

        Returns
        -------
        dict[str, dict[str, Any]]
            Detailed analysis


        Example:
        --------
        >>> test = IntExt(1024)
        >>> test.analyze()
        {
            'summary': {'number': 1024, 'length': 4, 'even': True, 'prime factor': [2^10], 'divisible': [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]},
            'convert': {'binary': '10000000000', 'octa': '2000', 'hex': '400', 'reverse': 4201, 'add to one': 7},
            'characteristic': {'prime': False, 'twisted prime': False, 'perfect': False, 'narcissistic': False, 'palindromic': False, 'palindromic prime': False}
        }
        """
        output = {
            "summary": {
                "number": self,
                "length": len(str(self)),
                "even": self.is_even(),
                "prime factor": self.prime_factor(short_form=short_form),  # type: ignore[call-overload]
                "divisible": self.divisible_list(),
            },
            "convert": {
                "binary": bin(self)[2:],
                "octa": oct(self)[2:],
                "hex": hex(self)[2:],
                # "hash": hash(self),
                "reverse": self.reverse(),
                "add to one": self.add_to_one_digit(),
            },
        }
        characteristic = {
            "prime": self.is_prime(),
            "twisted prime": self.is_twisted_prime(),
            "perfect": self.is_perfect(),
            "narcissistic": self.is_narcissistic(),
            "palindromic": self.is_palindromic(),
            "palindromic prime": self.is_palindromic_prime(),
        }
        if short_form:
            characteristic = DictBoolTrue(characteristic)

        output["characteristic"] = characteristic
        return output  # type: ignore

    @versionadded("5.1.0")
    def split(self, parts: int) -> list[int]:
        """
        Evenly split the number into ``parts`` parts

        Parameters
        ----------
        parts : int
            Split by how many parts

        Returns
        -------
        list[int]
            List of evenly splitted numbers


        Example:
        --------
        >>> IntExt(10).split(4)
        [2, 2, 3, 3]
        """
        p = max(1, parts)
        if p == 1:
            return [int(self)]

        quotient, remainder = divmod(self, p)
        return [quotient + (i >= (p - remainder)) for i in range(p)]


# Class
# ---------------------------------------------------------------------------
class IntBase(int, ABC):
    """
    A base class for creating custom integer-like types.
    Provides a hook for validation and extension.

    Usage:
    ------
    Use ``_validate(cls, value: int) -> None:`` classmethod to create custom validator
    """

    def __new__(cls, value: int = 0) -> Self:
        # Ensure the value is an integer
        if not isinstance(value, int):
            raise TypeError(f"{cls.__name__} must be initialized with an int, got {type(value).__name__}")
        # Optional validation hook
        cls._validate(value)
        return int.__new__(cls, value)

    @classmethod
    def _validate(cls, value: int) -> None:
        """Override in subclasses to add custom validation."""
        pass

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({int(self)})"


class PositiveInt(IntBase):
    """Only allows positive int"""

    @classmethod
    @override
    def _validate(cls, value: int) -> None:
        if value < 0:
            raise ValueError(f"{cls.__name__} must be non-negative")


class NegativeInt(IntBase):
    """Only allows negative int"""

    @classmethod
    @override
    def _validate(cls, value: int) -> None:
        if value >= 0:
            raise ValueError(f"{cls.__name__} must be non-positive")


class IntWithNote(IntBase):
    """Int with additional note"""

    def __new__(cls, value: int = 0) -> Self:
        ins = super().__new__(cls, value)
        ins.note = ""
        return ins

    def __repr__(self) -> str:
        if self.note == "":
            return f"{int(self)}"
            # return super().__repr__()
        return f"{self.__class__.__name__}({int(self)}, note={repr(self.note)})"

    def add_note(self, note: str) -> None:
        self.note = note



"""
Absfuyu: Shorten number
-----------------------
Short number base on suffixes

Version: 5.17.0
Date updated: 23/11/2025 (dd/mm/yyyy)
"""

from __future__ import annotations

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    "UnitSuffixFactory",
    "CommonUnitSuffixesFactory",
    "Decimal",
    "shorten_number",
    "Duration",
    "SupportDurationFormatPreset",
]


# Library
# ---------------------------------------------------------------------------
from collections.abc import Callable
from dataclasses import dataclass, field
from functools import wraps
from typing import Annotated, NamedTuple, ParamSpec, Protocol, Self, TypeVar

from absfuyu.core import versionadded

# Type
# ---------------------------------------------------------------------------
P = ParamSpec("P")  # Parameter type
N = TypeVar("N", int, float)  # Number type


# Class - Decimal
# ---------------------------------------------------------------------------
@versionadded("4.1.0")
class UnitSuffixFactory(NamedTuple):
    base: int
    short_name: list[str]
    full_name: list[str]


@versionadded("4.1.0")
class CommonUnitSuffixesFactory:
    NUMBER = UnitSuffixFactory(
        1000,
        [
            "",
            "K",
            "M",
            "B",
            "T",
            "Qa",
            "Qi",
            "Sx",
            "Sp",
            "Oc",
            "No",
            "Dc",
            "Ud",
            "Dd",
            "Td",
            "Qad",
            "Qid",
            "Sxd",
            "Spd",
            "Ocd",
            "Nod",
            "Vg",
            "Uvg",
            "Dvg",
            "Tvg",
            "Qavg",
            "Qivg",
            "Sxvg",
            "Spvg",
            "Ovg",
            "Nvg",
            "Tg",
            "Utg",
            "Dtg",
            "Ttg",
            "Qatg",
            "Qitg",
            "Sxtg",
            "Sptg",
            "Otg",
            "Ntg",
        ],
        [
            "",  # < Thousand
            "Thousand",
            "Million",
            "Billion",  # 1e9
            "Trillion",
            "Quadrillion",
            "Quintillion",
            "Sextillion",
            "Septillion",
            "Octillion",
            "Nonillion",
            "Decillion",  # 1e33
            "Undecillion",
            "Duodecillion",
            "Tredecillion",
            "Quattuordecillion",
            "Quindecillion",
            "Sexdecillion",
            "Septendecillion",
            "Octodecillion",
            "Novemdecillion",
            "Vigintillion",  #  1e63
            "Unvigintillion",
            "Duovigintillion",
            "Tresvigintillion",
            "Quattuorvigintillion",
            "Quinvigintillion",
            "Sesvigintillion",
            "Septemvigintillion",
            "Octovigintillion",
            "Novemvigintillion",
            "Trigintillion",  # 1e93
            "Untrigintillion",
            "Duotrigintillion",
            "Trestrigintillion",
            "Quattuortrigintillion",
            "Quintrigintillion",
            "Sestrigintillion",
            "Septentrigintillion",
            "Octotrigintillion",
            "Noventrigintillion",  #  1e120
        ],
    )
    DATA_SIZE = UnitSuffixFactory(
        1024,
        ["b", "Kb", "MB", "GB", "TB", "PB", "EB", "ZB", "YB", "BB"],
        [
            "byte",
            "kilobyte",
            "megabyte",
            "gigabyte",
            "terabyte",
            "petabyte",
            "exabyte",
            "zetabyte",
            "yottabyte",
            "brontobyte",
        ],
    )


@dataclass
@versionadded("4.1.0")
class Decimal:
    """
    Shorten large number

    Parameters
    ----------
    original_value : int | float
        Value to shorten

    base : int
        Short by base (must be > 0)

    suffixes : list[str]
        List of suffixes to use (ascending order)

    factory : UnitSuffixFactory | None
        ``UnitSuffixFactory`` to use
        (will overwrite ``base`` and ``suffixes``)

    suffix_full_name : bool
        Use suffix full name (available with ``UnitSuffixFactory``), by default ``False``

    Returns
    -------
    Decimal
        Decimal instance
    """

    original_value: int | float = field(repr=False)
    base: Annotated[int, "positive", "not_zero"] = field(repr=False, default=1000)
    suffixes: list[str] = field(repr=False, default_factory=list)
    factory: UnitSuffixFactory | None = field(repr=False, default=None)
    suffix_full_name: bool = field(repr=False, default=False)
    # Post init
    value: int | float = field(init=False)
    suffix: str = field(init=False)

    def __post_init__(self) -> None:
        self.base = max(1, self.base)  # Make sure that base >= 1
        self._get_factory()
        self.value, self.suffix = self._convert_decimal()

    def __str__(self) -> str:
        return self.to_text().strip()

    @classmethod
    def number(cls, value: int | float, suffix_full_name: bool = False) -> Self:
        """Decimal for normal large number"""
        return cls(
            value,
            factory=CommonUnitSuffixesFactory.NUMBER,
            suffix_full_name=suffix_full_name,
        )

    @classmethod
    def data_size(cls, value: int | float, suffix_full_name: bool = False) -> Self:
        """Decimal for data size"""
        return cls(
            value,
            factory=CommonUnitSuffixesFactory.DATA_SIZE,
            suffix_full_name=suffix_full_name,
        )

    @staticmethod
    def scientific_short(value: int | float) -> str:
        """Short number in scientific format"""
        return f"{value:.2e}"

    def _get_factory(self) -> None:
        if self.factory is not None:
            self.base = self.factory.base
            self.suffixes = self.factory.full_name if self.suffix_full_name else self.factory.short_name

    def _convert_decimal(self) -> tuple[float, str]:
        """Convert to smaller number"""
        suffix = self.suffixes[0] if len(self.suffixes) > 0 else ""
        unit = 1
        for i, suffix in enumerate(self.suffixes):
            unit = self.base**i
            if self.original_value < unit * self.base:
                break
        output = self.original_value / unit
        return output, suffix

    def to_text(self, decimal: int = 2, *, separator: str = " ", float_only: bool = True) -> str:
        """
        Convert to string

        Parameters
        ----------
        decimal : int, optional
            Round up to which decimal, by default ``2``

        separator : str, optional
            Character between value and suffix, by default ``" "``

        float_only : bool, optional
            Returns value as <float> instead of <int> when ``decimal = 0``, by default ``True``

        Returns
        -------
        str
            Decimal string
        """
        val = self.value.__round__(decimal)
        formatted_value = f"{val:,}"
        if not float_only and decimal == 0:
            formatted_value = f"{int(val):,}"
        return f"{formatted_value}{separator}{self.suffix}"


# Decorator
# ---------------------------------------------------------------------------
@versionadded("5.0.0")
def shorten_number(f: Callable[P, N]) -> Callable[P, Decimal]:
    """
    Shorten the number value by name

    Parameters
    ----------
    f : Callable[P, N]
        Function that return ``int`` or ``float``

    Returns
    -------
    Callable[P, Decimal]
        Function that return ``Decimal``


    Usage
    -----
    Use this as a decorator (``@shorten_number``)

    Example:
    --------
    >>> import random
    >>> @shorten_number
    >>> def big_num() -> int:
    ...     random.randint(100000000, 10000000000)
    >>> big_num()
    4.20 B
    """

    @wraps(f)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Decimal:
        value = Decimal.number(f(*args, **kwargs))
        return value

    return wrapper


# Class - Duration
# ---------------------------------------------------------------------------
# Format preset
class SupportDurationFormatPreset(Protocol):
    def __call__(self, duration: Duration, /) -> str: ...


@dataclass
@versionadded("5.16.0")
class Duration:
    """
    Convert duration in seconds to a more readable form. Eg: 3 mins 2 secs

    Parameters
    ----------
    total_seconds : int | float
        Seconds to convert to
    """

    total_seconds: int | float

    years: int = field(init=False)
    months: int = field(init=False)
    days: int = field(init=False)
    hours: int = field(init=False)
    minutes: int = field(init=False)
    seconds: int = field(init=False)

    _formats: dict[str, SupportDurationFormatPreset] = field(init=False)

    # Calculate duration
    def _calculate_duration(self) -> None:
        SEC_PER_MIN = 60
        SEC_PER_HOUR = 3600
        SEC_PER_DAY = 86400
        SEC_PER_MONTH = 30 * SEC_PER_DAY
        SEC_PER_YEAR = 365 * SEC_PER_DAY

        secs = self.total_seconds

        self.years, secs = divmod(secs, SEC_PER_YEAR)
        self.months, secs = divmod(secs, SEC_PER_MONTH)
        self.days, secs = divmod(secs, SEC_PER_DAY)
        self.hours, secs = divmod(secs, SEC_PER_HOUR)
        self.minutes, self.seconds = divmod(secs, SEC_PER_MIN)

    # Format handling
    def _init_format(self) -> None:

        def duration_compact_preset(duration: Self, /) -> str:
            """
            Example: "1y 2m 3d 4h 5m 6s"
            (fields = hidden when = 0).
            """
            parts = []
            if duration.years:
                parts.append(f"{duration.years}y")
            if duration.months:
                parts.append(f"{duration.months}m")
            if duration.days:
                parts.append(f"{duration.days}d")
            if duration.hours:
                parts.append(f"{duration.hours}h")
            if duration.minutes:
                parts.append(f"{duration.minutes}m")
            if duration.seconds:
                parts.append(f"{duration.seconds}s")
            return " ".join(parts) if parts else "0s"

        def duration_HMS_only_preset(duration: Self, /) -> str:
            """
            Example: "02:15:09" (HH:MM:SS only).
            """
            total = duration.total_seconds
            h, m = divmod(total, 3600)
            m, s = divmod(m, 60)
            return f"{h:02d}:{m:02d}:{s:02d}"

        def duration_digital_preset(duration: Self, /) -> str:
            """
            Examples:
            - If >= 1 day: "1d 02:03:04"
            - else: "02:03:04"
            """
            total = duration.total_seconds
            days, sec = divmod(total, 86400)
            h, sec = divmod(sec, 3600)
            m, s = divmod(sec, 60)

            if days:
                return f"{days}d {h:02d}:{m:02d}:{s:02d}"
            return f"{h:02d}:{m:02d}:{s:02d}"

        self._formats = {
            "compact": duration_compact_preset,
            "hms": duration_HMS_only_preset,
            "digital": duration_digital_preset,
        }

    @versionadded("5.17.0")
    def add_format(self, name: str, format_func: SupportDurationFormatPreset) -> None:
        """
        Add format style to Duration

        Parameters
        ----------
        name : str
            Name of the style (name will be lowercased)

        format_func : SupportDurationFormatPreset
            Format function
        """
        self._formats[name.lower().strip()] = format_func

    @property
    def available_formats(self) -> list[str]:
        """
        Available style format

        Returns
        -------
        list[str]
            All available style formats
        """
        return list(self._formats)

    def __format__(self, format_spec: str) -> str:
        """
        Change format of an object.

        Usage
        -----
        >>> print(f"{<object>:<format_spec>}")
        >>> print(<object>.__format__(<format_spec>))
        >>> print(format(<object>, <format_spec>))
        """

        func = self._formats.get(format_spec.lower().strip(), None)

        if func is None:
            return self.__str__()
        else:
            return func(self)

    # POST INIT
    def __post_init__(self) -> None:
        if not isinstance(self.total_seconds, (int, float)) or self.total_seconds < 0:
            raise ValueError("seconds must be a non-negative number")
        self._calculate_duration()
        self._init_format()

    def __str__(self) -> str:

        def _plural(n: int | float, word: str):
            return f"{n} {word}{'s' if n != 1 else ''}"

        parts = []
        if self.years:
            parts.append(_plural(self.years, "year"))
        if self.months:
            parts.append(_plural(self.months, "month"))
        if self.days:
            parts.append(_plural(self.days, "day"))
        if self.hours:
            parts.append(_plural(self.hours, "hour"))
        if self.minutes:
            parts.append(_plural(self.minutes, "minute"))
        if self.seconds:
            parts.append(_plural(self.seconds, "second"))
        return " ".join(parts) if parts else "0 second"

    # From other type of duration
    @classmethod
    def from_minute(cls, minutes: int | float) -> Self:
        return cls(minutes * 60)

    @classmethod
    def from_hour(cls, hours: int | float) -> Self:
        return cls(hours * 3600)

    @classmethod
    def from_day(cls, days: int | float) -> Self:
        return cls(days * 86400)

    @classmethod
    def from_month(cls, months: int | float) -> Self:
        return cls(months * 86400 * 30)

    @classmethod
    def from_year(cls, years: int | float) -> Self:
        return cls(years * 86400 * 365)



"""
Absfuyu: Image
--------------
Image related

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["ImgConverter"]


# Library
# ---------------------------------------------------------------------------
try:
    from PIL import Image
except ImportError:
    from subprocess import run

    from absfuyu.config import ABSFUYU_CONFIG

    if ABSFUYU_CONFIG._get_setting("auto-install-extra").value:  # type: ignore
        cmd = "python -m pip install -U absfuyu[pic]".split()
        run(cmd)
    else:
        raise SystemExit("This feature is in absfuyu[pic] package")  # noqa: B904

from absfuyu.extra.img.converter import ImgConverter



"""
Game: Tic Tac Toe
-----------------

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["TicTacToe", "GameMode"]


# Library
# ---------------------------------------------------------------------------
import random
import time
from typing import Literal, NamedTuple

from absfuyu.core import BaseClass, CLITextColor, versionadded

# Type
# ---------------------------------------------------------------------------
BoardGame = list[list[str]]


# Class
# ---------------------------------------------------------------------------
class Pos(NamedTuple):
    """Position"""

    row: int
    col: int


class GameMode:
    ONE_V_ONE = "1v1"
    ONE_V_BOT = "1v0"
    BOT_V_BOT = "0v0"


class GameStateResult(NamedTuple):
    key: str
    location: Literal["col", "row", "diag", "blank"]
    pos: int


@versionadded("3.3.3")
class TicTacToe(BaseClass):
    """Tic Tac Toe game"""

    def __init__(
        self,
        game_size: int = 3,
        *,
        x: str = "X",
        o: str = "O",
        blank: str = " ",
        position_split_symbol: str = ",",
        end_break_word: str = "END",
        welcome_message: bool = True,
    ) -> None:
        """
        :param game_size: Board size (Default: 3x3)
        :param x: X symbol
        :param o: O symbol
        :param blank: Blank symbol
        :param position_split_symbol: Position split symbol
        :param end_break_word: End break word
        :param welcome_message: Show welcome message (Default: `True`)
        """

        # Board size
        self.row_size = game_size
        self.col_size = game_size

        # Game setting
        self.X = x
        self.O = o  # noqa: E741
        self.BLANK = blank
        self.POS_SPLIT = position_split_symbol
        self.END_BREAK = end_break_word
        self.welcome_message = welcome_message

        # Init board
        self.board = self._gen_board()

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(game_size={self.row_size})"

    # Game
    def _gen_board(self) -> BoardGame:
        """
        Generate board game
        """
        board = [
            [self.BLANK for _ in range(self.row_size)] for _ in range(self.col_size)
        ]
        return board

    def _check_state(self) -> GameStateResult:
        """
        Check game winning state

        Returns
        -------
        dict[str, str | int]
            ``X`` | ``O`` | ``BLANK``
        """

        # Check rows
        for row in range(self.row_size):
            if len(set(self.board[row])) == 1:
                key = list(set(self.board[row]))[0]
                return GameStateResult(key, "row", row)
                # return {"key": key, "location": "row", "pos": row}  # modified

        # Check cols
        for col in range(self.col_size):
            temp = [self.board[row][col] for row in range(self.row_size)]
            if len(set(temp)) == 1:
                key = list(set(temp))[0]
                return GameStateResult(key, "col", col)
                # return {"key": key, "location": "col", "pos": col}  # modified

        # Check diagonal
        diag1 = [self.board[i][i] for i in range(len(self.board))]
        if len(set(diag1)) == 1:
            key = list(set(diag1))[0]
            return GameStateResult(key, "diag", 1)
            # return {"key": key, "location": "diag", "pos": 1}  # modified

        diag2 = [self.board[i][len(self.board) - i - 1] for i in range(len(self.board))]
        if len(set(diag2)) == 1:
            key = list(set(diag2))[0]
            return GameStateResult(key, "diag", 2)
            # return {"key": key, "location": "diag", "pos": 2}  # modified

        # Else
        return GameStateResult(self.BLANK, "blank", 0)
        # return {"key": self.BLANK}

    @staticmethod
    def _print_board(board: BoardGame) -> None:
        """
        Print Tic Tac Toe board
        """
        nrow, ncol = len(board), len(board[0])
        length = len(board)
        print(f"{'+---'*length}+")
        for row in range(nrow):
            for col in range(ncol):
                print(f"| {board[row][col]} ", end="")
            print(f"|\n{'+---'*length}+")

    def _win_hightlight(self) -> BoardGame:
        """
        Hight light the win by removing other placed key
        """

        # Get detailed information
        detail = self._check_state()
        loc = detail.location
        loc_line = detail.pos

        # Make new board
        board = self._gen_board()

        # Fill in the hightlighted content
        if loc.startswith("col"):
            for i in range(len(board)):
                board[i][loc_line] = detail.key
        elif loc.startswith("row"):
            for i in range(len(board)):
                board[loc_line][i] = detail.key
        else:
            if loc_line == 1:
                for i in range(len(board)):
                    board[i][i] = detail.key
            else:
                for i in range(len(board)):
                    board[i][len(board) - i - 1] = detail.key

        # Output
        return board

    def _is_blank(self, pos: Pos) -> bool:
        """Check if current pos is filled"""
        return self.board[pos.row][pos.col] == self.BLANK

    @staticmethod
    def _convert_bot_output(pos: Pos) -> Pos:
        """
        Turn to real pos by:

        - +1 to ``row`` and ``col``
        - convert into ``str``
        """
        return Pos(pos.row + 1, pos.col + 1)

    def _generate_random_move(self) -> Pos:
        """
        Generate a random move from board game
        """
        while True:
            output = Pos(
                random.randint(0, len(self.board) - 1),
                random.randint(0, len(self.board) - 1),
            )
            if self._is_blank(output):
                break
        return self._convert_bot_output(output)

    def play(
        self,
        game_mode: str = GameMode.ONE_V_BOT,
        *,
        bot_time: float = 0,
    ) -> None:
        """
        Play a game of Tic Tac Toe

        Parameters
        ----------
        game_mode : str
            Game mode

        bot_time : float
            Time sleep between each bot move (Default: ``0``)
        """
        # Init game
        filled = 0
        current_player = self.X
        state = self._check_state().key
        BOT = False
        BOT2 = False

        # Welcome message
        if self.welcome_message:
            print(
                f"""\
{CLITextColor.GREEN}Welcome to Tic Tac Toe!

{CLITextColor.YELLOW}Rules: Match lines vertically, horizontally or diagonally
{CLITextColor.YELLOW}{self.X} goes first, then {self.O}
{CLITextColor.RED}Type '{self.END_BREAK}' to end the game{CLITextColor.RESET}"""
            )

        # Check gamemode
        _game_mode = [
            "1v1",  # Player vs player
            "1v0",  # Player vs BOT
            "0v0",  # BOT vs BOT
        ]
        if game_mode not in _game_mode:
            game_mode = _game_mode[1]  # Force vs BOT
        if game_mode.startswith(GameMode.ONE_V_BOT):
            BOT = True
        if game_mode.startswith(GameMode.BOT_V_BOT):
            BOT = True
            BOT2 = True

        # Game
        self._print_board(self.board)

        place_pos = None
        while state == self.BLANK and filled < self.row_size**2:
            print(f"{CLITextColor.BLUE}{current_player}'s turn:{CLITextColor.RESET}")

            try:  # Error handling
                if (BOT and current_player == self.O) or BOT2:
                    move = self._generate_random_move()
                    str_move = f"{move.row}{self.POS_SPLIT}{move.col}"
                    move = str_move  # type: ignore
                else:
                    move = input(  # type: ignore
                        f"Place {CLITextColor.BLUE}{current_player}{CLITextColor.RESET} at {CLITextColor.BLUE}<row{self.POS_SPLIT}col>:{CLITextColor.RESET} "
                    )

                if move.upper() == self.END_BREAK:  # type: ignore # Failsafe
                    print(f"{CLITextColor.RED}Game ended{CLITextColor.RESET}")
                    break

                move = move.split(self.POS_SPLIT)  # type: ignore
                row = int(move[0])
                col = int(move[1])
                place_pos = Pos(row - 1, col - 1)

                if self._is_blank(place_pos):
                    self.board[place_pos.row][place_pos.col] = current_player
                    filled += 1

                else:  # User and BOT error
                    print(
                        f"{CLITextColor.RED}Invalid move, please try again{CLITextColor.RESET}"
                    )
                    continue

            except Exception:  # User error
                print(
                    f"{CLITextColor.RED}Invalid move, please try again{CLITextColor.RESET}"
                )
                continue

            state = self._check_state().key
            self._print_board(self.board)

            if state != self.BLANK:
                print(f"{CLITextColor.GREEN}{state} WON!{CLITextColor.RESET}")
                self._print_board(self._win_hightlight())

            # Change turn
            if BOT2:  # BOT delay
                time.sleep(bot_time)

            if current_player == self.X:
                current_player = self.O
            else:
                current_player = self.X

        if state == self.BLANK and filled == self.row_size**2:
            print(f"{CLITextColor.YELLOW}Draw Match!{CLITextColor.RESET}")



"""
Absfuyu: Rclone decrypt
-----------------------
Rclone decryptor

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["RcloneEncryptDecrypt", "DirectoryRcloneDEMixin"]


# Library
# ---------------------------------------------------------------------------
import os
import shutil
from pathlib import Path
from typing import Literal, Self

from rclone import Crypt

from absfuyu.core.baseclass import AutoREPRMixin
from absfuyu.logger import LogLevel, logger
from absfuyu.util.path import DirectoryBase


# Class
# ---------------------------------------------------------------------------
class RcloneEncryptDecrypt(AutoREPRMixin):
    """
    Rclone Decrypt/Encrypt Module
    """

    def __init__(self, crypt: Crypt) -> None:
        """
        This will encrypt/decrypt with rclone style

        Parameters
        ----------
        crypt : Crypt
            ``rclone.Crypt`` object | Encrypt/Decrypt engine
        """
        self._crypt = crypt

    @classmethod
    def from_passwd_salt(cls, passwd: str, salt: str | None = None) -> Self:
        """
        Create Rclone Decrypt/Encrypt object from password and salt

        Parameters
        ----------
        passwd : str
            Custom password

        salt : str | None, optional
            Custom salt, by default None

        Returns
        -------
        Self
            Rclone Decrypt/Encrypt object
        """
        if salt is None:
            crypt: Crypt = Crypt(passwd=passwd)
        else:
            crypt: Crypt = Crypt(passwd=passwd, salt=salt)
        return cls(crypt)

    # Support
    @staticmethod
    def _directory_validator(path: Path) -> Path:
        """Validate if directory exists, then return the path"""
        path = Path(path)
        if not path.exists():
            raise ValueError("Path does not exist")
        return path

    def _modify_name(self, name_to_modify: str, mode: Literal["encrypt", "decrypt"]) -> str:
        if mode == "decrypt":
            return self._crypt.Name.standard_decrypt(name_to_modify)
        if mode == "encrypt":
            return self._crypt.Name.standard_encrypt(name_to_modify)

    def _decrypt_encrypt_operation(
        self,
        mode: Literal["encrypt", "decrypt"],
        root_dir: Path | str,  # type: ignore
        delete_when_complete: bool = False,
    ) -> None:
        """
        Base operation to decrypt, encrypt

        Parameters
        ----------
        mode : Literal["encrypt", "decrypt"]
            Encrypt or decrypt

        root_dir : Path | str
            Directory location

        delete_when_complete : bool
            Delete directory when completed, defaults to False
        """

        logger.info(f"Begin operation: {mode.title()}")
        root_dir = Path(root_dir)
        FILE_MODE = False

        # Make Base folder
        if root_dir.is_file():
            # Get name of parent dir and make name
            temp = self._modify_name(root_dir.parent.name, "encrypt" if mode == "decrypt" else "decrypt")
            # Create sub dir
            base_dir = root_dir.parent.joinpath(temp)
            base_dir.mkdir(exist_ok=True, parents=True)
            # Move file to that dir
            new_path = root_dir.rename(base_dir.joinpath(root_dir.name))
            root_dir = base_dir  # Make root dir
            FILE_MODE = True
        else:
            root_dir: Path = self._directory_validator(root_dir)
            base_name: str = self._modify_name(root_dir.name, mode=mode)
            base_dir: Path = root_dir.parent.joinpath(base_name)
            base_dir.mkdir(exist_ok=True, parents=True)

        _ljust: int = 17  # Logger ljust value

        # Decrypt / Encrypt
        for path in root_dir.glob("**/*"):
            rel_path: Path = path.relative_to(root_dir)
            rel_path_splited: list[str] = str(rel_path).split(os.sep)
            rel_path_modified: list[str] = [self._modify_name(x, mode=mode) for x in rel_path_splited]
            new_path: Path = base_dir.joinpath(*rel_path_modified)

            if path.is_dir():
                logger.debug(f"{mode.title()}ing Dir:".ljust(_ljust) + f"{path}")
                new_path.mkdir(exist_ok=True, parents=True)
                logger.debug(f"{mode.title()}ed Dir:".ljust(_ljust) + f"{new_path}")
            else:
                logger.debug(f"{mode.title()}ing File:".ljust(_ljust) + f"{path}")
                if mode == "decrypt":
                    self._crypt.File.file_decrypt(path, new_path)  # type: ignore
                if mode == "encrypt":
                    self._crypt.File.file_encrypt(path, new_path)  # type: ignore
                logger.debug(f"{mode.title()}ed File:".ljust(_ljust) + f"{new_path}")
        logger.info(f"Operation: {mode.title()} COMPLETED")

        # File mode
        if FILE_MODE:
            new_path.rename(root_dir.parent.joinpath(new_path.name))  # type: ignore
            shutil.rmtree(root_dir)

        # Delete when completed
        if delete_when_complete:
            logger.debug(f"Deleting {root_dir}")
            shutil.rmtree(root_dir)
            logger.debug(f"Deleting {root_dir}...DONE")

    # Decrypt/Encrypt
    def decrypt(self, root_dir: Path, delete_when_complete: bool = False) -> None:
        """
        Decrypt encrypted directory.

        This will decrypt the directory itself and
        create a decrypted directory next to the original

        Parameters
        ----------
        root_dir : Path
            Directory location

        delete_when_complete : bool, optional
            Delete directory when completed, by default False
        """
        self._decrypt_encrypt_operation(mode="decrypt", root_dir=root_dir, delete_when_complete=delete_when_complete)

    def encrypt(self, root_dir: Path, delete_when_complete: bool = False) -> None:
        """
        Encrypt entire directory.

        This will encrypt the directory itself and
        create an encrypted directory next to the original

        Parameters
        ----------
        root_dir : Path
            Directory location

        delete_when_complete : bool, optional
            Delete directory when completed, by default False
        """
        self._decrypt_encrypt_operation(mode="encrypt", root_dir=root_dir, delete_when_complete=delete_when_complete)


class DirectoryRcloneDEMixin(DirectoryBase):
    """
    Directory - Rclone encrypt/decrypt

    Extension for ``absfuyu.util.path.Directory``

    - Decrypt
    - Encrypt
    """

    def rclone_encrypt(self, passwd: str, salt: str | None = None, delete_when_complete: bool = False) -> None:
        """
        Encrypt entire directory.

        This will encrypt the directory itself and
        create an encrypted directory next to the original

        Parameters
        ----------
        passwd : str
            Custom password

        salt : str | None, optional
            Custom salt, by default None

        delete_when_complete : bool, optional
            Delete directory when completed, by default False
        """
        engine = RcloneEncryptDecrypt.from_passwd_salt(passwd=passwd, salt=salt)
        engine.encrypt(self.source_path, delete_when_complete=delete_when_complete)

    def rclone_decrypt(self, passwd: str, salt: str | None = None, delete_when_complete: bool = False) -> None:
        """
        Decrypt encrypted directory.

        This will decrypt the directory itself and
        create a decrypted directory next to the original

        Parameters
        ----------
        passwd : str
            Custom password

        salt : str | None, optional
            Custom salt, by default None

        delete_when_complete : bool, optional
            Delete directory when completed, by default False
        """
        engine = RcloneEncryptDecrypt.from_passwd_salt(passwd=passwd, salt=salt)
        engine.decrypt(self.source_path, delete_when_complete=delete_when_complete)


# Run
# -----------------------------------------------------------------------------------------------
if __name__ == "__main__":
    logger.setLevel(LogLevel.DEBUG)



"""
Absfuyu: Data Analysis
----------------------
Matplotlib Helper

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["MatplotlibFormatString", "_PLTFormatString"]

# Library
# ---------------------------------------------------------------------------
import random
from collections.abc import Sequence
from itertools import product
from typing import ClassVar, Self

from absfuyu.core.baseclass import AutoREPRMixin, BaseClass


# Class
# ---------------------------------------------------------------------------
class _PLTFormatString(AutoREPRMixin):
    """
    Matplotlib format string

    Available format:
    - ``<marker><line><color>``
    - ``<color><marker><line>``
    """

    __slots__ = (
        "marker",
        "line_style",
        "color",
        "_marker_fullname",
        "_line_style_fullname",
        "_color_fullname",
    )

    def __init__(
        self,
        marker: str,
        line_style: str,
        color: str,
        *,
        marker_fullname: str | None = None,
        line_style_fullname: str | None = None,
        color_fullname: str | None = None,
    ) -> None:
        """
        Matplotlib format string

        Parameters
        ----------
        marker : str
            Maker

        line_style : str
            Line style

        color : str
            Color

        marker_fullname : str | None, optional
            Maker fullname, by default ``None``

        line_style_fullname : str | None, optional
            Line style fullname, by default ``None``

        color_fullname : str | None, optional
            Color fullname, by default ``None``
        """

        self.marker = marker
        self.line_style = line_style
        self.color = color

        self._marker_fullname = marker_fullname
        self._line_style_fullname = line_style_fullname
        self._color_fullname = color_fullname

    def __str__(self) -> str:
        return self.fstr

    def __format__(self, format_spec: str) -> str:
        if format_spec.lower() == "full":
            if (
                self._marker_fullname is None
                or self._line_style_fullname is None
                or self._color_fullname is None
            ):
                return self.__str__()
            clsname = self.__class__.__name__
            return (
                f"{clsname}(marker={repr(self._marker_fullname)}"
                f", line_style={repr(self._line_style_fullname)}"
                f", color={repr(self._color_fullname)})"
            )
        return super().__format__(format_spec)

    @property
    def fstr(self) -> str:
        """Format string"""
        return f"{self.marker}{self.line_style}{self.color}"

    @property
    def alternate(self) -> str:
        """Alternative version of format string"""
        return f"{self.color}{self.marker}{self.line_style}"

    @classmethod
    def _make(cls, iterable: Sequence[str]) -> Self:
        if len(iterable) not in (3, 6):
            raise ValueError("iterable must have a length of 3 or 6")
        try:  # Len 6
            return cls(
                marker=iterable[0],
                line_style=iterable[1],
                color=iterable[2],
                marker_fullname=iterable[3],
                line_style_fullname=iterable[4],
                color_fullname=iterable[5],
            )
        except IndexError:  # Len 3
            return cls(iterable[0], iterable[1], iterable[2])


class MatplotlibFormatString(BaseClass):
    """
    Matplotlib format string

    Available format:
    - ``<marker><line><color>``
    - ``<color><marker><line>``
    """

    MARKER_DATA: ClassVar[dict[str, str]] = {
        ".": "point marker",
        ",": "pixel marker",
        "o": "circle marker",
        "v": "triangle_down marker",
        "^": "triangle_up marker",
        "<": "triangle_left marker",
        ">": "triangle_right marker",
        "1": "tri_down marker",
        "2": "tri_up marker",
        "3": "tri_left marker",
        "4": "tri_right marker",
        "8": "octagon marker",
        "s": "square marker",
        "p": "pentagon marker",
        "P": "plus (filled) marker",
        "*": "star marker",
        "h": "hexagon1 marker",
        "H": "hexagon2 marker",
        "+": "plus marker",
        "x": "x marker",
        "X": "x (filled) marker",
        "D": "diamond marker",
        "d": "thin_diamond marker",
        "|": "vline marker",
        "_": "hline marker",
    }
    LINE_STYLE_DATA: ClassVar[dict[str, str]] = {
        "-": "solid line style",
        "--": "dashed line style",
        "-.": "dash-dot line style",
        ":": "dotted line style",
    }
    COLOR_DATA: ClassVar[dict[str, str]] = {
        "b": "blue",
        "g": "green",
        "r": "red",
        "c": "cyan",
        "m": "magenta",
        "y": "yellow",
        "k": "black",
        "w": "white",
    }

    @classmethod
    def all_format_string(cls) -> list[_PLTFormatString]:
        # This return full list without full name
        # return [
        #     _PLTFormatString._make(x)
        #     for x in product(cls.MARKER_DATA, cls.LINE_STYLE_DATA, cls.COLOR_DATA)
        # ]

        # This return full list with full name
        def convert(
            x: tuple[tuple[str, str], tuple[str, str], tuple[str, str]],
        ) -> tuple[str, str, str, str, str, str]:
            return (x[0][0], x[1][0], x[2][0], x[0][1], x[1][1], x[2][1])

        return [
            _PLTFormatString._make(convert(x))
            for x in product(
                cls.MARKER_DATA.items(),
                cls.LINE_STYLE_DATA.items(),
                cls.COLOR_DATA.items(),
            )
        ]

    @classmethod
    def get_random(cls) -> _PLTFormatString:
        """
        Get a random format string

        Returns
        -------
        str
            Random format string
        """
        random_fmtstr = random.choice(cls.all_format_string())
        return random_fmtstr



"""
Absfuyu: Path
-------------
Path related

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)

Feature:
--------
- Directory
- SaveFileAs
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    # Main
    "DirectoryBase",
    "Directory",
    "SaveFileAs",
    # Mixin
    "DirectoryInfoMixin",
    "DirectoryBasicOperationMixin",
    "DirectoryArchiverMixin",
    "DirectoryOrganizerMixin",
    "DirectoryTreeMixin",
    "DirectorySelectMixin",
    # Support
    "FileOrFolderWithModificationTime",
    "DirectoryInfo",
]


# Library
# ---------------------------------------------------------------------------
import os
import re
import shutil
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import Any, ClassVar, Literal, NamedTuple

from absfuyu.core.baseclass import BaseClass
from absfuyu.core.decorator import add_subclass_methods_decorator
from absfuyu.core.docstring import deprecated, versionadded, versionchanged
from absfuyu.logger import logger

# Template
# ---------------------------------------------------------------------------
ORGANIZE_TEMPLATE: dict[str, list[str]] = {
    "Code": [
        ".ps1",
        ".py",
        ".rs",
        ".js",
        ".c",
        ".h",
        ".cpp",
        ".cs",
        ".r",
        ".cmd",
        ".bat",
        ".lua",
    ],
    "Comic": [".cbz", ".cbr", ".cb7", ".cbt", ".cba"],
    "Compressed": [
        ".7z",
        ".zip",
        ".rar",
        ".apk",
        ".cab",
        ".tar",
        ".tgz",
        ".txz",
        ".bz2",
        ".gz",
        ".lz",
        ".lz4",
        ".lzma",
        ".xz",
        ".zipx",
        ".zst",
    ],
    "Documents": [".docx", ".doc", ".xls", ".xlsx", ".ppt", ".pptx", ".pdf", ".txt"],
    "Ebook": [
        ".epub",
        ".mobi",
        ".prc",
        ".lrf",
        ".lrx",
        ".pdb",
        ".azw",
        ".azw3",
        ".kf8",
        ".kfx",
        ".opf",
    ],
    "Music": [".mp3", ".flac", ".wav", ".m4a", ".wma", ".aac", ".alac", ".aiff"],
    "OS": [".iso", ".dmg", ".wim"],
    "Pictures": [
        ".jpg",
        ".jpeg",
        ".png",
        ".apng",
        ".avif",
        ".bmp",
        ".gif",
        ".jfif",
        ".pjpeg",
        ".pjp",
        ".svg",
        ".ico",
        ".cur",
        ".tif",
        ".tiff",
        ".webp",
    ],
    "Programs": [".exe", ".msi"],
    "Video": [
        ".3g2",
        ".3gp",
        ".avi",
        ".flv",
        ".m2ts",
        ".m4v",
        ".mkv",
        ".mov",
        ".mp4",
        ".mpeg",
        ".mpv",
        ".mts",
        ".ts",
        ".vob",
        ".webm",
    ],
    "Raw pictures": [
        ".3fr",
        ".ari",
        ".arw",
        ".bay",
        ".braw",
        ".crw",
        ".cr2",
        ".cr3",
        ".cap",
        ".data",
        ".dcs",
        ".dcr",
        ".dng",
        ".drf",
        ".eip",
        ".erf",
        ".fff",
        ".gpr",
        ".iiq",
        ".k25",
        ".kdc",
        ".mdc",
        ".mef",
        ".mos",
        ".mrw",
        ".nef",
        ".nrw",
        ".obm",
        ".orf",
        ".pef",
        ".ptx",
        ".pxn",
        ".r3d",
        ".raf",
        ".raw",
        ".rwl",
        ".rw2",
        ".rwz",
        ".sr2",
        ".srf",
        ".srw",
        ".tif",
        ".x3f",
    ],
}


# Support Class
# ---------------------------------------------------------------------------
@versionadded("3.3.0")
class FileOrFolderWithModificationTime(NamedTuple):
    """
    File or Folder with modification time

    :param path: Original path
    :param modification_time: Modification time
    """

    path: Path
    modification_time: datetime


@deprecated(
    "5.1.0", reason="Support for ``DirectoryInfoMixin`` which is also deprecated"
)
@versionadded("3.3.0")
class DirectoryInfo(NamedTuple):
    """
    Information of a directory
    """

    creation_time: datetime
    modification_time: datetime


# Class - Directory
# ---------------------------------------------------------------------------
@add_subclass_methods_decorator
class DirectoryBase(BaseClass):
    """
    Directory - Base

    Parameters
    ----------
    source_path : str | Path
        Source folder

    create_if_not_exist : bool
        Create directory when not exist,
        by default ``False``
    """

    # Custom attribute
    _METHOD_INCLUDE: ClassVar[bool] = True  # Include in DIR_METHODS
    SUBCLASS_METHODS: ClassVar[dict[str, list[str]]] = {}

    def __init__(
        self,
        source_path: str | Path,
        create_if_not_exist: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        source_path : str | Path
            Source folder

        create_if_not_exist : bool
            Create directory when not exist,
            by default ``False``
        """
        self.source_path = Path(source_path)
        if not self.source_path.exists():
            if create_if_not_exist:
                self.source_path.mkdir(exist_ok=True, parents=True)
            else:
                raise FileNotFoundError("Directory not existed")


class DirectoryInfoMixin(DirectoryBase):
    """
    Directory - Info

    - Quick info
    """

    @deprecated("5.1.0", reason="Not efficient")
    @versionadded("3.3.0")
    def quick_info(self) -> DirectoryInfo:
        """
        Quick information about this Directory

        Returns
        -------
        DirectoryInfo
            DirectoryInfo
        """
        source_stat: os.stat_result = self.source_path.stat()
        out = DirectoryInfo(
            creation_time=datetime.fromtimestamp(source_stat.st_ctime),
            modification_time=datetime.fromtimestamp(source_stat.st_mtime),
        )
        return out


class DirectoryBasicOperationMixin(DirectoryBase):
    """
    Directory - Basic operation

    - Rename
    - Copy
    - Move
    - Delete
    """

    # Rename
    def rename(self, new_name: str) -> None:
        """
        Rename directory

        Parameters
        ----------
        new_name : str
            Name only (not the entire path)
        """
        try:
            logger.debug(f"Renaming to {new_name}...")
            self.source_path.rename(self.source_path.with_name(new_name))
            logger.debug(f"Renaming to {new_name}...DONE")
        except Exception as e:
            logger.error(e)
        # return self.source_path

    # Copy
    def copy(self, dst: Path) -> None:
        """
        Copy entire directory

        Parameters
        ----------
        dst : Path
            Destination
        """
        logger.debug(f"Copying to {dst}...")
        try:
            try:
                shutil.copytree(self.source_path, Path(dst), dirs_exist_ok=True)
            except Exception:
                shutil.copytree(self.source_path, Path(dst))
            logger.debug(f"Copying to {dst}...DONE")
        except Exception as e:
            logger.error(e)

    # Move
    def move(self, dst: Path, content_only: bool = False) -> None:
        """
        Move entire directory

        Parameters
        ----------
        dst : Path
            Destination

        content_only : bool
            Only move content inside the folder (Default: ``False``; Move entire folder)
        """
        try:
            logger.debug(f"Moving to {dst}...")
            if content_only:
                for x in self.source_path.iterdir():
                    shutil.move(x, Path(dst))
            else:
                shutil.move(self.source_path, Path(dst))
            logger.debug(f"Moving to {dst}...DONE")

        except OSError as e:  # File already exists
            logger.error(e)
            logger.debug("Overwriting file...")
            if content_only:
                for x in self.source_path.iterdir():
                    shutil.move(x, Path(dst).joinpath(x.name))
            else:
                shutil.move(self.source_path, Path(dst))
            logger.debug("Overwriting file...DONE")

    # Delete folder
    def _mtime_folder(self) -> list[FileOrFolderWithModificationTime]:
        """
        Get modification time of file/folder (first level only)
        """
        return [
            FileOrFolderWithModificationTime(
                path, datetime.fromtimestamp(path.stat().st_mtime)
            )
            for path in self.source_path.glob("*")
        ]

    @staticmethod
    def _delete_files(list_of_files: list[Path]) -> None:
        """
        Delete files/folders
        """
        for x in list_of_files:
            x = Path(x).absolute()
            logger.debug(f"Removing {x}...")
            try:
                if x.is_dir():
                    shutil.rmtree(x)
                else:
                    x.unlink()
                logger.debug(f"Removing {x}...SUCCEED")
            except Exception:
                logger.error(f"Removing {x}...FAILED")

    @staticmethod
    def _date_filter(
        value: FileOrFolderWithModificationTime,
        period: Literal["Y", "M", "D"] = "Y",
    ) -> bool:
        """
        Filter out file with current Year|Month|Day
        """
        data = {
            "Y": value.modification_time.year,
            "M": value.modification_time.month,
            "D": value.modification_time.day,
        }
        now = datetime.now()
        ntime = {"Y": now.year, "M": now.month, "D": now.day}
        return data[period] != ntime[period]

    def delete(
        self,
        entire: bool = False,
        *,
        based_on_time: bool = False,
        keep: Literal["Y", "M", "D"] = "Y",
    ) -> None:
        """
        Deletes everything

        Parameters
        ----------
        entire : bool
            | ``True``: Deletes the folder itself
            | ``False``: Deletes content inside only
            | (Default: ``False``)

        based_on_time : bool
            | ``True``: Deletes everything except ``keep`` period
            | ``False``: Works normal
            | (Default: ``False``)

        keep : Literal["Y", "M", "D"]
            Delete all file except current ``Year`` | ``Month`` | ``Day``
        """
        try:
            logger.info(f"Removing {self.source_path}...")

            if entire:
                shutil.rmtree(self.source_path)
            else:
                if based_on_time:
                    filter_func = partial(self._date_filter, period=keep)
                    # self._delete_files([x[0] for x in filter(filter_func, self._mtime_folder())])
                    self._delete_files(
                        [x.path for x in filter(filter_func, self._mtime_folder())]
                    )
                else:
                    self._delete_files(
                        map(lambda x: x.path, self._mtime_folder())  # type: ignore
                    )

            logger.info(f"Removing {self.source_path}...SUCCEED")
        except Exception as e:
            logger.error(f"Removing {self.source_path}...FAILED\n{e}")


class DirectoryArchiverMixin(DirectoryBase):
    """
    Directory - Archiver/Compress

    - Compress
    - Decompress
    - Register extra zip format <staticmethod>
    """

    @versionchanged("5.1.0", reason="Update funcionality (new parameter)")
    def compress(
        self,
        format: Literal["zip", "tar", "gztar", "bztar", "xztar"] = "zip",
        delete_after_compress: bool = False,
        move_inside: bool = True,
    ) -> Path | None:
        """
        Compress the directory (Default: Create ``.zip`` file)

        Parameters
        ----------
        format : Literal["zip", "tar", "gztar", "bztar", "xztar"], optional
            By default ``"zip"``
            - ``zip``: ZIP file (if the ``zlib`` module is available).
            - ``tar``: Uncompressed tar file. Uses POSIX.1-2001 pax format for new archives.
            - ``gztar``: gzip'ed tar-file (if the ``zlib`` module is available).
            - ``bztar``: bzip2'ed tar-file (if the ``bz2`` module is available).
            - ``xztar``: xz'ed tar-file (if the ``lzma`` module is available).

        delete_after_compress : bool, optional
            Delete directory after compress, by default ``False``

        move_inside : bool, optional
            Move the commpressed file inside the directory,
            by default ``True``

        Returns
        -------
        Path
            Compressed path

        None
            When fail to compress
        """
        logger.debug(f"Zipping {self.source_path}...")
        try:
            # Zip
            # zip_name = self.source_path.parent.joinpath(self.source_path.name).__str__()
            # shutil.make_archive(zip_name, format=format, root_dir=self.source_path)
            zip_path = shutil.make_archive(
                self.source_path.__str__(), format=format, root_dir=self.source_path
            )
            logger.debug(f"Zipping {self.source_path}...DONE")
            logger.debug(f"Path: {zip_path}")

            # Del
            if delete_after_compress:
                move_inside = False
                shutil.rmtree(self.source_path)

            # Move
            if move_inside:
                zf = Path(zip_path)
                _move_path = self.source_path.joinpath(zf.name)
                if _move_path.exists():
                    _move_path.unlink(missing_ok=True)
                _move = zf.rename(_move_path)
                return _move

            return Path(zip_path)
        except (FileExistsError, OSError) as e:
            logger.error(f"Zipping {self.source_path}...FAILED\n{e}")
            return None

    @staticmethod
    @versionadded("5.1.0")
    def register_extra_zip_format() -> None:
        """This register extra extension for zipfile"""
        extra_extension = [".zip", ".cbz"]
        shutil.unregister_unpack_format("zip")
        shutil.register_unpack_format(
            "zip",
            extra_extension,
            shutil._unpack_zipfile,  # type: ignore
            description="ZIP file",
        )

    @versionadded("5.1.0")
    def decompress(
        self,
        format: Literal["zip", "tar", "gztar", "bztar", "xztar"] | None = None,
        delete_after_done: bool = False,
    ) -> None:
        """
        Decompress compressed file in directory (first level only)

        Parameters
        ----------
        format : Literal["zip", "tar", "gztar", "bztar", "xztar"] | None, optional
            By default ``None``
            - ``zip``: ZIP file (if the ``zlib`` module is available).
            - ``tar``: Uncompressed tar file. Uses POSIX.1-2001 pax format for new archives.
            - ``gztar``: gzip'ed tar-file (if the ``zlib`` module is available).
            - ``bztar``: bzip2'ed tar-file (if the ``bz2`` module is available).
            - ``xztar``: xz'ed tar-file (if the ``lzma`` module is available).

        delete_after_done : bool, optional
            Delete compressed file when extracted, by default ``False``
        """
        # Register extra extension
        self.register_extra_zip_format()

        # Decompress first level only
        for path in self.source_path.glob("*"):
            try:
                shutil.unpack_archive(
                    path, path.parent.joinpath(path.stem), format=format
                )
                if delete_after_done and path.is_file():
                    path.unlink(missing_ok=True)
            except OSError:
                continue


class DirectoryOrganizerMixin(DirectoryBase):
    """
    Directory - File organizer

    - Organize
    """

    @versionadded("5.3.0")
    def organize(self, dirtemplate: dict[str, list[str]] | None = None) -> None:
        """
        Organize a directory.

        Parameters
        ----------
        dirtemplate : dict[str, Collection[str]] | None, optional
            | Template to move file to, by default ``None``.
            | Example: {"Documents": [".txt", ".pdf", ...]}
        """
        if dirtemplate is None:
            template = ORGANIZE_TEMPLATE
        else:
            template = dirtemplate

        other_dir = self.source_path.joinpath("Others")
        other_dir.mkdir(parents=True, exist_ok=True)

        for path in self.source_path.iterdir():
            if path.is_dir():
                continue

            for dir_name, suffixes in template.items():
                if path.suffix.lower() in suffixes:
                    move_path = self.source_path.joinpath(dir_name)
                    move_path.mkdir(parents=True, exist_ok=True)
                    path.rename(move_path.joinpath(path.name))
                    break
            else:
                path.rename(other_dir.joinpath(path.name))


class DirectoryTreeMixin(DirectoryBase):
    # Directory structure
    def _list_dir(self, *ignore: str) -> list[Path]:
        """
        List all directories and files

        Parameters
        ----------
        ignore : str
            List of pattern to ignore. Example: "__pycache__", ".pyc"
        """
        logger.debug(f"Base folder: {self.source_path.name}")

        list_of_path = self.source_path.glob("**/*")

        # No ignore rules
        if len(ignore) == 0:  # No ignore pattern
            return [path.relative_to(self.source_path) for path in list_of_path]

        # With ignore rules
        # ignore_pattern = "|".join(ignore)
        ignore_pattern = re.compile("|".join(ignore))
        logger.debug(f"Ignore pattern: {ignore_pattern}")
        return [
            path.relative_to(self.source_path)
            for path in list_of_path
            if re.search(ignore_pattern, path.name) is None
        ]

    @staticmethod
    @versionadded("3.3.0")
    def _split_dir(list_of_path: list[Path]) -> list[list[str]]:
        """
        Split pathname by ``os.sep``

        Parameters
        ----------
        list_of_path : list[Path]
            List of Path

        Returns
        -------
        list[list[str]]
            List of splitted dir


        Example:
        --------
        >>> test = [Path(test_root / test_not_root), ...]
        >>> Directory._split_dir(test)
        [[test_root, test_not_root], [...]...]
        """

        return sorted([str(path).split(os.sep) for path in list_of_path])

    def _separate_dir_and_files(
        self,
        list_of_path: list[Path],
        *,
        tab_symbol: str | None = None,
        sub_dir_symbol: str | None = None,
    ) -> list[str]:
        """
        Separate dir and file and transform into folder structure

        Parameters
        ----------
        list_of_path : list[Path]
            List of paths

        tab_symbol : str | None
            Tab symbol
            (Default: ``"\\t"``)

        sub_dir_symbol : str | None
            Sub-directory symbol
            (Default: ``"|-- "``)

        Returns
        -------
        list[str]
            Folder structure ready to print
        """
        # Check for tab and sub-dir symbol
        if tab_symbol is None:
            tab_symbol = "\t"
        if sub_dir_symbol is None:
            sub_dir_symbol = "|-- "

        temp: list[list[str]] = self._split_dir(list_of_path)

        return [  # Returns n-tab space with sub-dir-symbol for the last item in x
            f"{tab_symbol * (len(x) - 1)}{sub_dir_symbol}{x[-1]}" for x in temp
        ]

    def list_structure(self, *ignore: str) -> str:
        """
        List folder structure

        Parameters
        ----------
        ignore : str
            Tuple contains patterns to ignore

        Returns
        -------
        str
            Directory structure


        Example (For typical python library):
        -------------------------------------
        >>> test = Directory(<source path>)
        >>> test.list_structure(
                "__pycache__",
                ".pyc",
                "__init__",
                "__main__",
            )
        ...
        """
        temp: list[Path] = self._list_dir(*ignore)
        out: list[str] = self._separate_dir_and_files(temp)
        return "\n".join(out)  # Join the list

    def list_structure_pkg(self) -> str:
        """
        List folder structure of a typical python package

        Returns
        -------
        str
            Directory structure
        """
        return self.list_structure("__pycache__", ".pyc")


@versionadded("5.6.0")
class DirectorySelectMixin(DirectoryBase):
    """
    Directory - File select

    - Select all
    """

    def select_all(self, *file_type: str, recursive: bool = False) -> list[Path]:
        """
        Select all files

        Parameters
        ----------
        file_type : str
            File suffix to select

        recursive : bool, optional
            Include sub directories, by default ``False``

        Returns
        -------
        list[Path]
            Selected file paths
        """
        pattern = "**/*" if recursive else "*"
        paths = [
            x
            for x in self.source_path.glob(pattern)
            if x.is_file() and x.suffix.lower() in map(lambda x: x.lower(), file_type)
        ]
        return paths


class Directory(
    DirectoryTreeMixin,
    DirectoryOrganizerMixin,
    DirectoryArchiverMixin,
    DirectoryBasicOperationMixin,
    DirectorySelectMixin,
    DirectoryInfoMixin,
):
    """
    Some shortcuts for directory

    Parameters
    ----------
    source_path : str | Path
        Source folder

    create_if_not_exist : bool
        Create directory when not exist,
        by default ``False``


    Example:
    --------
    >>> # For a list of method
    >>> Directory.SUBCLASS_METHODS
    """

    pass


# Class - SaveFileAs
# ---------------------------------------------------------------------------
class SaveFileAs:
    """
    File as multiple file type
    """

    def __init__(self, data: Any, *, encoding: str | None = "utf-8") -> None:
        """
        :param encoding: Default: utf-8
        """
        self.data = data
        self.encoding = encoding

    def __str__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __repr__(self) -> str:
        return self.__str__()

    def to_txt(self, path: str | Path) -> None:
        """
        Save as ``.txt`` file

        Parameters
        ----------
        path : Path
            Save location
        """
        with open(path, "w", encoding=self.encoding) as file:
            file.writelines(self.data)

    # def to_pickle(self, path: Union[str, Path]) -> None:
    #     """
    #     Save as .pickle file

    #     :param path: Save location
    #     """
    #     from absfuyu.util.pkl import Pickler
    #     Pickler.save(path, self.data)

    # def to_json(self, path: Union[str, Path]) -> None:
    #     """
    #     Save as .json file

    #     :param path: Save location
    #     """
    #     from absfuyu.util.json_method import JsonFile
    #     temp = JsonFile(path, sort_keys=False)
    #     temp.save_json()



"""
Absfuyu: Data Analysis
----------------------
Data Analyst

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    "MatplotlibFormatString",
    "DADF",
    # Function
    "custom_pandas_settings",
    "reset_custom_pandas_settings",
]


# Library
# ---------------------------------------------------------------------------
DA_MODE = False

try:
    import numpy as np
    import openpyxl
    import pandas as pd
    import xlsxwriter
except ImportError:
    from subprocess import run

    from absfuyu.config import ABSFUYU_CONFIG

    if ABSFUYU_CONFIG._get_setting("auto-install-extra").value:  # type: ignore
        cmd = "python -m pip install -U absfuyu[full]".split()
        run(cmd)
    else:
        raise SystemExit("This feature is in absfuyu[full] package")  # noqa: B904
else:
    DA_MODE = True

from absfuyu.extra.da.dadf import DADF
from absfuyu.extra.da.mplt import MatplotlibFormatString


# Function
# ---------------------------------------------------------------------------
def custom_pandas_settings(*, show_all_rows: bool = False) -> None:
    """
    Custom pandas settings. Currently only show all cols/rows

    Parameters
    ----------
    show_all_rows : bool, optional
        Show all rows, by default False
    """
    # Shows all columns
    pd.set_option("display.max_columns", None)  # type: ignore

    if show_all_rows:
        # (optional) also show all rows if needed
        pd.set_option("display.max_rows", None)  # type: ignore


def reset_custom_pandas_settings() -> None:
    """
    Reset custom pandas settings
    """
    settings = ["display.max_columns", "display.max_rows"]
    for x in settings:
        pd.reset_option(x)  # type: ignore



"""
Test: Util - Performance

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

import pytest

from absfuyu.util.performance import retry


@pytest.mark.abs_util
class TestUtilPerformance:
    """absfuyu.util.performance"""

    # retry
    def test_retry_invalid_parameters(self):
        with pytest.raises(
            ValueError, match="retries must be >= 1, delay must be >= 0"
        ):

            @retry(retries=0)
            def invalid_function():
                pass

        with pytest.raises(
            ValueError, match="retries must be >= 1, delay must be >= 0"
        ):

            @retry(retries=3, delay=-1)
            def invalid_function():
                pass

    def test_retry_success(self) -> None:
        @retry(retries=3, delay=0.1)
        def success_function():
            return "Success"

        result = success_function()
        assert result == "Success"

    @pytest.mark.xfail
    def test_retry_failure(self) -> None:
        with pytest.raises(Exception, match="Function error") as excinfo:

            @retry(retries=3, delay=0.1)
            def failing_function():
                raise Exception("Function error")

            failing_function()
        assert str(excinfo.value)

    def test_retry_with_valid_retries(self) -> None:
        @retry(retries=2, delay=0.1)
        def intermittent_failure():
            if intermittent_failure.call_count < 1:
                intermittent_failure.call_count += 1
                raise Exception("Temporary failure")
            return "Success"

        intermittent_failure.call_count = 0  # Initialize call count

        result = intermittent_failure()

        assert result == "Success"



"""
Absfuyu: Data Analysis
----------------------
DF Function

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["equalize_df", "compare_2_list", "rename_with_dict"]


# Library
# ---------------------------------------------------------------------------
from itertools import chain

import numpy as np
import pandas as pd


# Function
# ---------------------------------------------------------------------------
def equalize_df(data: dict[str, list], fillna=np.nan) -> dict[str, list]:
    """
    Make all list in dict have equal length to make pd.DataFrame

    :param data: `dict` data that ready for `pd.DataFrame`
    :param fillna: Fill N/A value (Default: `np.nan`)
    """
    max_len = max(map(len, data.values()))
    for _, v in data.items():
        if len(v) < max_len:
            missings = max_len - len(v)
            for _ in range(missings):
                v.append(fillna)
    return data


def compare_2_list(*arr) -> pd.DataFrame:
    """
    Compare 2 lists then create DataFrame
    to see which items are missing

    Parameters
    ----------
    arr : list
        List

    Returns
    -------
    DataFrame
        Compare result
    """
    # Setup
    col_name = "list"
    arr = [sorted(x) for x in arr]  # type: ignore # map(sorted, arr)

    # Total array
    tarr = sorted(list(set(chain.from_iterable(arr))))
    # max_len = len(tarr)

    # Temp dataset
    temp_dict = {"base": tarr}
    for idx, x in enumerate(arr):
        name = f"{col_name}{idx}"

        # convert list
        temp = [item if item in x else np.nan for item in tarr]

        temp_dict.setdefault(name, temp)

    df = pd.DataFrame(temp_dict)
    df["Compare"] = np.where(
        df[f"{col_name}0"].apply(lambda x: str(x).lower())
        == df[f"{col_name}1"].apply(lambda x: str(x).lower()),
        df[f"{col_name}0"],  # Value when True
        np.nan,  # Value when False
    )
    return df


def rename_with_dict(df: pd.DataFrame, col: str, rename_dict: dict) -> pd.DataFrame:
    """
    Version: 2.0.0

    :param df: DataFrame
    :param col: Column name
    :param rename_dict: Rename dictionary
    """

    name = f"{col}_filtered"
    df[name] = df[col]
    rename_val = list(rename_dict.keys())
    df[name] = df[name].apply(lambda x: "Other" if x in rename_val else x)
    return df
